#ifndef __berw_h
#define __berw_h

static inline guint32
be_read32 (int fd)
{
	union {
	 	guint8 buf[4];
		guint32 word;
	} u;

	read(fd, u.buf, 4);

	return ntohl(u.word);
}

static inline guint32
be_read24 (int fd)
{
	guint32 word;
	guint8 buf[3];


	read(fd, buf, 3);
	word = (buf[0] << 16) | (buf[1] << 8) | buf[2];

	if (buf[0] > 0x7f)
		word |= (0xff << 24);

	return word;
}

static inline guint16
be_read16 (int fd)
{
	union {
		guint16 word;
		guint8 buf[2];
	} u;

	read(fd, u.buf, 2);

	return ntohs(u.word);
}

static inline ssize_t
be_write16 (int fd, guint16 word)
{
	union {
		guint8 buf[2];
		guint16 word;
	} u;

	u.word = htons(word);

	return write(fd, u.buf, 2);
}

static inline ssize_t
be_write24 (int fd, guint32 word)
{
 	guint8 buf[3];


	buf[2] = word & 0xff;
	buf[1] = (word & 0xff00) >> 8;
	buf[0] = (word & 0xff0000) >> 16;
	return write(fd, buf, 3);
}

static inline ssize_t
be_write32 (int fd, guint32 word)
{
	union {
		guint32 word;
		guint8 buf[4];
	} u;


	u.word = htonl(word);
	return write(fd, u.buf, 4);
}

static inline guint32 _e_int32(unsigned char **base)
{
	unsigned char *b = *base;
	guint32 ret = (*b++) << 24;
	ret |= (*b++) << 16;
	ret |= (*b++) << 8;
	ret |= *b++;
	*base = b;
	return ret;
}

static inline guint32 _e_int24(unsigned char **base)
{
	unsigned char *b = *base;
	guint32 ret = (*b++) << 16;
	ret |= (*b++) << 8;
	ret |= *b++;
	*base = b;
	return ret;
}

static inline guint16 _e_int16(unsigned char **base)
{
	unsigned char *b = *base;
	guint16 ret = (*b++) << 8;
	ret |= *b++;
	*base = b;
	return ret;
}

#define e_int32(foo) _e_int32(&foo)
#define e_int24(foo) _e_int24(&foo)
#define e_int16(foo) _e_int16(&foo)
#endif
