#ifndef __hxd_HX_H
#define __hxd_HX_H 1
#include "hotline.h"

#ifndef MAXPATHLEN
#ifdef PATH_MAX
#define MAXPATHLEN PATH_MAX
#else
#define MAXPATHLEN 4095
#endif
#endif

#define WHITE           "\033[0;37m"
#define WHITE_BOLD      "\033[1;37m"
#define RED             "\033[0;31m"
#define RED_BOLD        "\033[1;31m"

#define XFER_GET	0
#define XFER_PUT	1
#define COMPLETE_NONE	0
#define COMPLETE_EXPAND	1
#define COMPLETE_LS_R	2
#define COMPLETE_GET_R	3

#define MAX_CONN 1

#ifdef HAVE_DCGETTEXT
#include <libintl.h>
#define _(string) dgettext (PACKAGE, string)
#else
#define _(string) (string)
#endif

typedef struct {
	int xsize, ysize;
	int xpos, ypos;
	unsigned char open:1;
	unsigned char init:1;
} Window_Geo;

struct gtkhx_prefs {
	int num_tracker;
	int num_icons;
	int trans_xtext;
#ifdef USE_GDK_PIXBUF
	int tint_red, tint_blue, tint_green;
#endif
	char *sound_path;
	char *auto_reply_msg;
	char *font;
	char *download_path;
	char **tracker;
	char *tracker_str;
	char **icon;
	char *icon_str;
	char *snd_cmd;
	int xbuf_max;

	struct {
		GtkWidget *options_window;
		GtkWidget *nick;
		GtkWidget *icon;
		GtkWidget *showjoin;
		GtkWidget *tracker_list;
		GtkWidget *showback;
		GtkWidget *queuedl;
		GtkWidget *icon_path;
		GtkWidget *sound_path;
		GtkWidget *downloadpath;
		GtkWidget *auto_reply;
		GtkWidget *auto_reply_msg;
		GtkWidget *font;
		GtkWidget *use_fontset;
		GtkWidget *trans_xtext;
		GtkWidget *timestamp;
		GtkWidget *snd_cmd;
		GtkWidget *word_wrap;
		GtkWidget *file_samewin;
		GtkWidget *track_case;
		GtkWidget *xbuf_max;
		GtkWidget *old_nickcompletion;
	} gtk;

	struct {
		Window_Geo chat;
		Window_Geo news;
		Window_Geo tool;
		Window_Geo tasks;
		Window_Geo users;
	} geo;

	unsigned char queuedl:1;
	unsigned char showjoin:1;
	unsigned char showback:1;
	unsigned char auto_reply:1;
	unsigned char use_fontset:1;
	unsigned char timestamp:1;
	unsigned char word_wrap:1;
	unsigned char file_samewin:1;
	unsigned char track_case:1;
	unsigned char old_nickcompletion:1;
};

extern struct gtkhx_prefs gtkhx_prefs;

struct msgwin {
	struct msgwin *next, *prev;
	guint32 uid;
	char *name;
	GtkWidget *outputbuf;
	GtkWidget *inputbuf;
	GtkWidget *vscroll;
	GtkWidget *window;
};

struct timer {
	struct timer *next, *prev;
	guint id;
	int (*fn)();
	void *ptr;
};

#include "macres.h"

struct ifn {
	char **files;
	macres_file **cicns;
	unsigned int n;
};

struct gtkhx_chat {
	struct gtkhx_chat *next, *prev;
	GtkWidget *window;
	GtkWidget *vscroll;
	GtkWidget *output;
	GtkWidget *input;
	GtkWidget *subject;
	GtkWidget *userlist;
	guint32 cid;
	struct chat *chat;
	void *chat_history;
};

struct hx_sounds {
	struct {
		GtkWidget *on;
		GtkWidget *inv;
		GtkWidget *chat;
		GtkWidget *err;
		GtkWidget *file;
		GtkWidget *join;
		GtkWidget *login;
		GtkWidget *msg;
		GtkWidget *news;
		GtkWidget *part;
	} gtk;
	unsigned char on:1;
	unsigned char invite:1, chat:1, error:1, file:1, join:1, login:1, msg:1, news:1, part:1;
};

struct date_time {
  guint16 base_year;
  guint16 pad;
  guint32 seconds;
};

struct news_post {
	char *buf;
	struct news_item *item;
};

struct news_item {
	int postid,parentid;
	char *sender;
	char *subject;
	struct date_time date;
	int partcount;
	int size;
	struct {
		int size;
		char *mime_type;
	} *parts;
	GtkCTreeNode *node;
	struct news_group *group;
};

struct news_group {
	int post_count;
	char *path;
	struct news_item *posts;
};

struct folder_item {
	char *name;
	guint16 icon;
	int type;
};

struct news_folder {
	struct folder_item **entry;
	char *path;
	int num_entries;
};

struct uesp_fn {
	void *uesp;
	void (*fn)(void *, const char *, const char *, const char *, const struct hl_access_bits *);
};


#if MAXPATHLEN > 4095
#undef MAXPATHLEN
#define MAXPATHLEN 4095
#endif

struct task {

        struct task *next, *prev;
        guint32 trans;
        guint32 pos, len;
        void *data;

        char *str;
        void *ptr;
        void (*rcv)();
};

typedef struct _chat_invite {
	guint32 cid, uid;
	struct _chat_invite *next, *prev;
} chat_invite;


struct hx_user {
	struct hx_user *next, *prev;
	guint32 uid;
	guint16 icon;
	guint16 color;
	guint8 name[32];
	unsigned char ignore:1;
};

struct chat {
	struct chat *next, *prev;
	guint32 cid;
	guint32 nusers;
	struct hx_user __user_list;
	struct hx_user *user_list;
	struct hx_user *user_tail;
	char subject[256];
};


typedef struct _session {
	GtkWidget *toolbar_window;
	GtkWidget *news_window;
	GtkWidget *chat_window;
	GtkWidget *tasks_window;
	GtkWidget *users_window;

	GtkWidget *users_list;

	GtkWidget *news_text;
	GtkWidget *postButton;
	GtkWidget *reloadButton;

	GtkWidget *agreementwin;

	struct gtkhx_chat *gchat_list;

	struct gtask *gtask_list;
	GtkWidget *gtklist, *gtask_scroll;

	struct gnews_folder *gfnews_list;

	chat_invite *chat_invite_list;

	struct gfile_list *gfile_list;

	struct msgwin *msg_list;

	struct gnews_catalog *gcnews_list;

	struct task __task_list;
	struct task *task_list, *task_tail;

	struct hx_user __user_list;
	struct hx_user *user_list, *user_tail;

	struct chat *chat_front, *chat_tail, *chat_list;
	struct chat __chat_list;

	struct cached_filelist *cfl_list;

	struct htlc_conn htlc;

	unsigned char connected:1;
} session;

extern session sessions[MAX_CONN];

extern session *sess_from_htlc(struct htlc_conn *htlc);

extern char last_msg_nick[32];
#define INFOPREFIX	" \00310[\003\00303hx\003\00310]\003 "
extern char *g_user_colors[4];
#define USERCOLORS	(g_user_colors)
#define DEFAULT_USER_COLORS	{GREEN_BOLD, GREEN, RED_BOLD, RED}

extern char *colorstr (guint16 color);


extern void hotline_client_input (struct htlc_conn *htlc, char *str, guint32 cid, guint16 style);



extern void hx_htlc_close (struct htlc_conn *htlc);
extern void hx_connect (struct htlc_conn *htlc, const char *serverstr, guint16 port,
			const char *login, const char *pass);




extern void hx_tracker_list (session *sess, char *addrstr, guint16 port);



extern void hx_quit (void);

extern guint8 *news_buf;
extern size_t news_len;

extern char *expand_user_name (char *name, unsigned int nlen);
extern guint32 cmd_arg(int argc, char *str);

#ifdef USE_DEBUG
#define debug(args...) printf(args)
#else
#define debug(args...)
#endif


extern void chrexpand (char *str, int len);

#if !defined(__va_copy)
#define __va_copy(_dst, _src) ((_dst) = (_src))
#endif

static inline void
strip_ansi (char *buf, int len)
{
	register char *p, *end = buf + len;

	for (p = buf; p < end; p++)
		if (*p < 31 && *p > 13 && *p != 15 && *p != 22)
				*p = (*p & 127) | 64;
}


struct cached_filelist {
	struct cached_filelist *next, *prev;
	char *path;
	struct hl_filelist_hdr *fh;
	guint32 fhlen;
	unsigned char completing:2;
	char **filter_argv;
};

struct output_functions {
	void (*init)(int argc, char **argv);
	void (*loop)(void);
	void (*clear)(struct htlc_conn *htlc, guint32 cid, int subj);
	void (*chat)(struct htlc_conn *htlc, guint32 cid, char *chat, guint16 len);
	void (*msg)(char *name, guint32 uid, char *buf);
	void (*agreement)(session *sess, const char *agreement, guint16 len);
	void (*news_file)(struct htlc_conn *htlc, char *news, guint16 len);
	void (*news_post)(struct htlc_conn *htlc, char *news, guint16 len);
	void (*user_info)(guint32 uid, const char *nam, const char *info, guint16 len);
	void (*file_info) (char *name, char *creator, char *type, char *comments, char *modified, char *created, guint32 size);
	void (*user_create)(struct htlc_conn *htlc, struct chat *chat, struct hx_user *user, const char *nam, guint16 icon, guint16 color);
	void (*user_delete)(struct htlc_conn *htlc, struct chat *chat, struct hx_user *user);
	void (*user_change)(struct htlc_conn *htlc, struct chat *chat, struct hx_user *user, const char *nam, guint16 icon, guint16 color);
	void (*user_list)(struct htlc_conn *htlc);
	void (*users_clear)(struct htlc_conn *htlc, struct chat *chat);
	void (*file_list)(struct cached_filelist *cfl, struct hl_filelist_hdr *fh, void *data);
	void (*file_update)(session *sess, struct htxf_conn *htxf);
	void (*tracker_server_create)(struct in_addr addr, guint16 port, guint16 nusers, const char *nam, const char *desc, int total);
	void (*task_update)(session *sess, struct task *tsk);
	void (*news_folder)(struct news_folder *folder);
	void (*news_catalog)(struct news_group *group);
	void (*news_thread)(struct news_post *post);
	void (*chat_subject)(struct htlc_conn *htlc, guint32 cid, char *buf);
	void (*chat_invitation)(struct htlc_conn *htlc, guint32 cid, char *name);
	void (*xfer_queue)(struct htxf_conn *htxf);
	void (*tracker_clear)(void);
};

extern struct output_functions hx_output;
extern void play_sound(int sound);
extern void timer_add_secs (time_t secs, int (*fn)(), void *ptr);
extern void timer_delete_ptr (void *ptr);

extern void changetitlesconnected();
extern void changetitlesdisconnected();
extern void error_dialog(char *title, char *message);

#endif /* ndef __hxd_HX_H */
