#ifndef _HXD_H
#define _HXD_H

#if !defined(__GNUC__) || defined(__STRICT_ANSI__) || defined(__APPLE_CC__)
#define __attribute__(x)
#endif

#define HOSTLEN 256

#define MAX_HOTLINE_PACKET_LEN 0x100000
#define UNKNOWN_TYPECREA	"TEXTR*ch"

#ifndef MAXPATHLEN
#ifdef PATH_MAX
#define MAXPATHLEN PATH_MAX
#else
#define MAXPATHLEN 4095
#endif
#endif

#if MAXPATHLEN > 4095
#undef MAXPATHLEN
#define MAXPATHLEN 4095
#endif

#include <time.h>
#include <sys/time.h>

struct qbuf {
	guint32 pos, len;
	guint8 *buf;
};

struct htlc_chat {
	struct htlc_chat *next, *prev;
	guint32 ref;
	guint8 subject[256];
	fd_set fds;
	guint32 nusers;
};

struct extra_access_bits {
	guint32 chat_private:1,
		  msg:1,
		  user_getlist:1,
		  file_list:1,
		  file_getinfo:1,
		  file_hash:1,
		  can_login:1,
		  is_server:1,
		  user_visibility:1,
		  user_color:1,
		  can_spam:1,
		  __reserved:21;
};

struct htlc_conn;

struct htxf_conn {
	guint32 data_size, data_pos, rsrc_size, rsrc_pos;
	guint32 total_size, total_pos;
	guint32 ref;   /* xfer id */
	guint8 gone;
	guint8 type;
	guint32 queue; /* position in server queue */
	int fd;
	int pipe;
	pid_t pid;
#ifdef USE_IPV6
	struct addrinfo *listen_addr;
#else
	struct sockaddr_in listen_addr;
#endif
	struct htlc_conn *htlc;
	char path[MAXPATHLEN];
	char remotepath[MAXPATHLEN];
	struct qbuf in;
	char **filter_argv;
	struct timeval start;
	struct {
		guint32 retry:1, reserved:31;
	} opt;
};

struct htlc_conn {
	struct htlc_conn *next, *prev;
	void (*rcv)(struct htlc_conn *);
	void (*real_rcv)(struct htlc_conn *);
	struct qbuf in, out;
#ifdef USE_IPV6
	struct addrinfo *addr;
#else
	struct sockaddr_in addr;
#endif
	int fd;
	guint32 trans;
	guint32 chattrans;
	guint32 icon;
	guint16 uid;
	guint16 version;

	struct {
		guint32 visible:1, reserved:31;
	} flags;


	struct hl_access_bits access;
	struct extra_access_bits access_extra;
	guint16 current_uid;
	guint8 userid[513];
	guint8 name[32];
	guint8 login[32];
	char rootdir[MAXPATHLEN];

	unsigned int gdk_input:1;
	unsigned int color:7;
};

struct hxd_file {
	union {
		void *ptr;
		struct htlc_conn *htlc;
		struct htrk_conn *htrk;
		struct htxf_conn *htxf;
	} conn;
	void (*ready_read)(int fd);
	void (*ready_write)(int fd);
};

extern struct hxd_file *hxd_files;

extern int hxd_open_max;

extern void hxd_fd_add (int fd);
extern void hxd_fd_del (int fd);
extern void hxd_fd_set (int fd, int rw);
extern void hxd_fd_clr (int fd, int rw);
#define FDR	1
#define FDW	2

extern char **hxd_environ;



#if !defined(HAVE_INET_NTOA_R)
extern int inet_ntoa_r (struct in_addr in, char *buf, size_t buflen);
#endif

#ifndef HAVE_LOCALTIME_R
extern struct tm *localtime_r (const time_t *t, struct tm *tm);
#endif

#if !defined(HAVE_SNPRINTF) || defined(__hpux__)
extern int snprintf (char *str, size_t count, const char *fmt, ...);
#endif

#if !defined(HAVE_VSNPRINTF) || defined(__hpux__)
extern int vsnprintf (char *str, size_t count, const char *fmt, va_list ap);
#endif

#ifndef HAVE_BASENAME
extern char *basename (const char *path);
#endif

#ifndef RETSIGTYPE
#define RETSIGTYPE void
#endif

extern int fd_blocking (int fd, int on);
extern int fd_closeonexec (int fd, int on);
extern int fd_lock_write (int fd);
extern void timer_add (struct timeval *tv, void (*fn)(), void *ptr);

extern void qbuf_set (struct qbuf *q, guint32 pos, guint32 len);
extern void qbuf_add (struct qbuf *q, void *buf, guint32 len);

extern void htlc_close (struct htlc_conn *htlc);

extern void start_ident (struct htlc_conn *htlc);
extern int check_banlist (struct htlc_conn *htlc);
extern void addto_banlist (const char *name, const char *login, const char *user, const char *address, const char *message);

extern void snd_strerror (struct htlc_conn *htlc, int err);

extern void hlwrite (struct htlc_conn *htlc, guint32 type, guint32 flag, int hc, ...);
extern void hl_code (void *__dst, const void *__src, size_t len);
#define hl_decode(d,s,l) hl_code(d,s,l)
#define hl_encode(d,s,l) hl_code(d,s,l)

extern int account_read (const char *login, char *password, char *name, struct hl_access_bits *acc);
extern int account_write (const char *login, const char *password, const char *name, const struct hl_access_bits *acc);
extern int account_delete (const char *login);
extern int account_trusted (const char *login, const char *userid, const char *addr);
extern void account_getconf (struct htlc_conn *htlc);
extern void account_get_access_extra (struct htlc_conn *htlc);

extern void news_send_file (struct htlc_conn *htlc);
extern void news_save_post (char *newsfile, guint8 *buf, guint16 len);
extern void agreement_send_file (struct htlc_conn *htlc);

extern int chat_isset (struct htlc_conn *, struct htlc_chat *);
extern struct htlc_chat *chat_lookup_ref (guint32);
extern void chat_remove_from_all (struct htlc_conn *htlc);

extern void command_chat (struct htlc_conn *htlc, char *chatbuf);

#define atou32(_str) ((guint32)strtoul(_str, 0, 0))
#define atou16(_str) ((guint16)strtoul(_str, 0, 0))

#define DHARGS_BUF_SIZE	(0x10000 + SIZEOF_HL_DATA_HDR)
extern guint8 *_dhargs_buf;

#if 1
/* data must be accessed from locations that are aligned on multiples of the data size */
#define dh_start(_htlc)													\
{															\
	struct hl_data_hdr *dh;												\
	guint32 _pos, _max;												\
	memcpy(_dhargs_buf, &((_htlc)->in.buf[SIZEOF_HL_HDR]), SIZEOF_HL_DATA_HDR);				\
	dh = (struct hl_data_hdr *)_dhargs_buf;										\
	for (_pos = SIZEOF_HL_HDR, _max = (_htlc)->in.pos;								\
	     _pos + SIZEOF_HL_DATA_HDR <= _max && (guint16)ntohs(dh->len) <= (_max - _pos) - SIZEOF_HL_DATA_HDR;	\
	     memcpy(_dhargs_buf, &((_htlc)->in.buf[_pos]), SIZEOF_HL_DATA_HDR)) {					\
		memcpy(dh->data, &((_htlc)->in.buf[_pos + SIZEOF_HL_DATA_HDR]), ntohs(dh->len));			\
		_pos += SIZEOF_HL_DATA_HDR + ntohs(dh->len);

#define L32NTOH(_word, _addr) \
	do { guint32 _x; memcpy(&_x, (_addr), 4); _word = ntohl(_x); } while (0)
#define S32HTON(_word, _addr) \
	do { guint32 _x; _x = htonl(_word); memcpy((_addr), &_x, 4); } while (0)
#define L16NTOH(_word, _addr) \
	do { guint16 _x; memcpy(&_x, (_addr), 2); _word = ntohs(_x); } while (0)
#define S16HTON(_word, _addr) \
	do { guint16 _x; _x = htons(_word); memcpy((_addr), &_x, 2); } while (0)
#else
#define dh_start(_htlc)													\
{															\
	struct hl_data_hdr *dh = (struct hl_data_hdr *)(&((_htlc)->in.buf[SIZEOF_HL_HDR]));				\
	guint32 _pos, _max;												\
	for (_pos = SIZEOF_HL_HDR, _max = (_htlc)->in.pos;								\
	     _pos + SIZEOF_HL_DATA_HDR < _max && (guint16)ntohs(dh->len) <= (_max - _pos) - SIZEOF_HL_DATA_HDR;	\
	     _pos += SIZEOF_HL_DATA_HDR + ntohs(dh->len),								\
	     dh = (struct hl_data_hdr *)(((guint8 *)dh) + SIZEOF_HL_DATA_HDR + ntohs(dh->len))) {

#define L32NTOH(_word, _addr) \
	_word = ntohl(*((guint32 *)_addr))
#define S32HTON(_word, _addr) \
	*((guint32 *)_addr) = htonl(_word)
#define L16NTOH(_word, _addr) \
	_word = ntohs(*((guint16 *)_addr))
#define S16HTON(_word, _addr) \
	*((guint16 *)_addr) = htons(_word)
#endif

#define dh_getint(_word)						\
do {									\
	if (ntohs(dh->len) == 4)					\
		_word = (guint32)ntohl(*((guint32 *)dh->data));	\
	else /* if (ntohs(dh->len) == 2) */				\
		_word = (guint32)ntohs(*((guint16 *)dh->data));	\
} while (0)

#define dh_end()	\
	}		\
}

#define X2X(_ptr, _len, _x1, _x2) \
do {						\
	char *_p = _ptr, *_end = _ptr + _len;	\
	for ( ; _p < _end; _p++)		\
		if (*_p == _x1)			\
			*_p = _x2;		\
} while (0)

#define CR2LF(_ptr, _len)	X2X(_ptr, _len, '\r', '\n')
#define LF2CR(_ptr, _len)	X2X(_ptr, _len, '\n', '\r')

#endif /* ndef _HXD_H */
