#include <stdio.h>
#include "config.h"
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <gtk/gtk.h>
#include <pthread.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/select.h>
#include <time.h>
#include <netinet/in.h>
#include "hotline.h"
#include "hxd.h"
#include "hx.h"
#include "rcv.h"
#include "gtkthreads.h"
#include "gtkutil.h"
#include "chat.h"
#include "tasks.h"
#include "users.h"

char *server_addr;
#ifdef USE_IPV6
guint16 server_port;
#endif

pthread_t conn_tid;
int connected;

int fd_blocking (int fd, int on)
{
	int x;

#if defined(_POSIX_SOURCE) || !defined(FIONBIO)
#if !defined(O_NONBLOCK)
# if defined(O_NDELAY)
#  define O_NONBLOCK O_NDELAY
# endif
#endif
	if ((x = fcntl(fd, F_GETFL, 0)) == -1)
		return -1;
	if (on)
		x &= ~O_NONBLOCK;
	else
		x |= O_NONBLOCK;

	return fcntl(fd, F_SETFL, x);
#else
	x = !on;

	return ioctl(fd, FIONBIO, &x);
#endif
}

int
fd_closeonexec (int fd, int on)
{
	int x;

	if ((x = fcntl(fd, F_GETFD, 0)) == -1)
		return -1;
	if (on)
		x &= ~FD_CLOEXEC;
	else
		x |= FD_CLOEXEC;

	return fcntl(fd, F_SETFD, x);
}

int
fd_lock_write (int fd)
{
	struct flock lk;

	lk.l_type = F_WRLCK;
	lk.l_start = 0;
	lk.l_whence = SEEK_SET;
	lk.l_len = 0;
	lk.l_pid = getpid();

	return fcntl(fd, F_SETLK, &lk);
}

void
hx_htlc_close (struct htlc_conn *htlc)
{
	int fd = htlc->fd;
	struct chat *chat, *cnext;
	struct hx_user *user, *unext;
	struct task *tsk, *tsknext;
	char buf[HOSTLEN];

	session *sess = sess_from_htlc(htlc);

	if(conn_tid) {
		pthread_cancel(conn_tid);
		conn_tid = 0;
	}
#ifdef USE_IPV6
	getnameinfo(htlc->addr->ai_addr, htlc->addr->ai_addrlen, buf,

				sizeof(buf), NULL, 0, NI_NUMERICHOST);
#else
	inet_ntop(AF_INET, &htlc->addr.sin_addr, buf, sizeof(buf));
#endif
	hx_printf_prefix(htlc, 0, INFOPREFIX, "%s: %s\n", buf,

					 _("connection closed"));
	connected = 0;
	setbtns(sess, 0);
	set_disconnect_btn(sess, 0);
	set_status_bar(0);
	close_connected_windows (sess);
	close(fd);
	changetitlesdisconnected(sess);
	hxd_fd_clr(fd, FDR|FDW);
	if (htlc->in.buf) {
		g_free(htlc->in.buf);
		htlc->in.buf = NULL;
	}
	if (htlc->out.buf) {
		g_free(htlc->out.buf);
		htlc->out.buf = NULL;
	}
	memset(&hxd_files[fd], 0, sizeof(struct hxd_file));
	htlc->fd = 0;
	htlc->uid = 0;
	htlc->color = 0;
	htlc->gdk_input = 0;
	memset(htlc->login, 0, sizeof(htlc->login));


	for (chat = sess->chat_list; chat; chat = cnext) {
		cnext = chat->next;
		hx_output.users_clear(htlc, chat);
		for (user = chat->user_list->next; user; user = unext) {
			unext = user->next;
			hx_user_delete(&chat->user_tail, user);
		}
		if (chat != sess->chat_list)
			chat_delete(sess, chat);
	}
	memset(sess->chat_list, 0, sizeof(struct chat));
	sess->chat_list->user_list = sess->chat_list->user_tail =

		&sess->chat_list->__user_list;

	for (tsk = sess->task_list->next; tsk; tsk = tsknext) {
		tsknext = tsk->next;
		task_delete(sess, tsk);
	}


#ifdef USE_IPV6
	freeaddrinfo(htlc->addr);
	htlc->addr = NULL;
#endif
}


static void htlc_read (int fd)
{
	ssize_t r;
	struct htlc_conn *htlc = hxd_files[fd].conn.htlc;


	r = read(fd, &htlc->in.buf[htlc->in.pos], htlc->in.len);
	if (r == 0 || (r < 0 && errno != EWOULDBLOCK && errno != EINTR)) {
		hx_htlc_close(htlc);
	}
	else {
		htlc->in.pos += r;
		htlc->in.len -= r;
		if (htlc->in.pos >= SIZEOF_HL_HDR) {
			struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
			if (ntohl(h->type) == HTLS_HDR_TASK) {
				struct task *tsk = task_with_trans(&sessions[0],

												   ntohl(h->trans));
				if(ntohs(h->flag) & 1) {
					goto bleh;
				}
				tsk->pos = htlc->in.pos;
				tsk->len = htlc->in.len;
				hx_output.task_update(&sessions[0], tsk);
			}
		}
		if (!htlc->in.len) {
		  bleh:
			if (htlc->rcv) {
				if (htlc->rcv == hx_rcv_hdr) {
					hx_rcv_hdr(htlc);
				}
				else {
					htlc->rcv(htlc);
					goto reset;
				}
			}
			else {
			  reset:
				htlc->rcv = hx_rcv_hdr;
				qbuf_set(&htlc->in, 0, SIZEOF_HL_HDR);
			}
		}
	}
}

static void htlc_write (int fd)
{
	ssize_t r;
	struct htlc_conn *htlc = hxd_files[fd].conn.htlc;

	r = write(fd, &htlc->out.buf[htlc->out.pos], htlc->out.len);
	if (r == 0 || (r < 0 && errno != EWOULDBLOCK && errno != EINTR)) {
		hx_htlc_close(htlc);
	}
	else {
		htlc->out.pos += r;
		htlc->out.len -= r;
		if (!htlc->out.len) {
			htlc->out.pos = 0;
			htlc->out.len = 0;
			hxd_fd_clr(fd, FDW);
		}
	}
}

struct connect_data {
	struct htlc_conn *htlc;
	char *serverstr, *login, *pass;
	guint16 port;
	int s;
};


static void hx_thread_connect (void *arg)
{
	int s;
	struct connect_data *cdata = arg;
	char enclogin[64], encpass[64];
	guint16 icon16;
	guint16 llen, plen;
	char *serverstr = cdata->serverstr;
	struct htlc_conn *htlc = cdata->htlc;
	char *login = cdata->login;
	char *pass = cdata->pass;
	guint16 port = cdata->port;
	session *sess = sess_from_htlc(htlc);
#ifdef USE_IPV6
	char portstr[HOSTLEN];
	struct addrinfo *he;
	struct addrinfo hints;
	int error;

	server_port = port;
	sprintf(portstr, "%u", port);

	memset(&hints, 0, sizeof(struct addrinfo));
	hints.ai_family = PF_UNSPEC;
	hints.ai_flags = AI_CANONNAME;
	hints.ai_socktype = SOCK_STREAM;
#else
	struct sockaddr_in saddr;

	memset(&saddr, 0, sizeof(struct sockaddr_in));
	saddr.sin_port = htons(port);
	saddr.sin_family = AF_INET;
#endif


#ifndef USE_DEBUG
# ifdef USING_DARWIN
	conn_tid = getpid();
# endif
#endif
	
	htlc->gdk_input = 0;
	server_addr = g_strdup_printf("%s:%u", serverstr, port);

	gtk_threads_enter();
	hx_printf_prefix(htlc, 0, INFOPREFIX, "%s %s\n", _("connecting to"),

					 server_addr);
	conn_task_update(sess, 0);
	set_status_bar(-1);
	set_disconnect_btn(sess, 1);
	gtk_threads_leave();

#ifndef USE_IPV6
	if(!(inet_pton(AF_INET, serverstr, &saddr.sin_addr))) {
		struct hostent *he;

		if((he = gethostbyname(serverstr))) {
			size_t len = (unsigned)he->h_length > sizeof(struct in_addr)
				? sizeof(struct in_addr) : he->h_length;
			memcpy(&saddr.sin_addr, he->h_addr, len);
		}
		else {
#else
		if((error = getaddrinfo(serverstr, portstr, &hints, &he))) {
#endif
			gtk_threads_enter();

#ifdef USE_IPV6
			hx_printf_prefix(htlc, 0, INFOPREFIX, "%s: %s\n", serverstr,

							 gai_strerror(error));
#else
# ifdef HAVE_HSTRERROR
			hx_printf_prefix(htlc, 0, INFOPREFIX, "DNS lookup for %s failed: %s\n", serverstr, hstrerror(h_errno));
# else
			hx_printf_prefix(htlc, 0, INFOPREFIX, "DNS lookup for %s failed\n",
							 serverstr);
# endif
#endif


			setbtns(sess, 0);
			set_status_bar(0);
			set_disconnect_btn(sess, 0);
			conn_task_update(sess, 2);
			gtk_threads_leave();
			return;
		}
#ifndef USE_IPV6
		}
#endif


#ifdef USE_IPV6
		s = socket(he->ai_family, SOCK_STREAM, IPPROTO_TCP);
#else
		s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
#endif

    if (s < 0) {
		gtk_threads_enter();
		hx_printf_prefix(htlc, 0, INFOPREFIX, "socket: %s\n", strerror(errno));
		setbtns(sess, 0);
		set_status_bar(0);
		set_disconnect_btn(sess, 0);
		conn_task_update(sess, 2);
		gtk_threads_leave();
		return;
	}

	if (s >= hxd_open_max) {
		gtk_threads_enter();
		hx_printf_prefix(htlc, 0, INFOPREFIX, "%s:%d: %d >= hxd_open_max (%d)",
						 __FILE__, __LINE__, s, hxd_open_max);
		setbtns(sess, 0);
		set_status_bar(0);
		set_disconnect_btn(sess, 0);
		conn_task_update(sess, 2);
		gtk_threads_leave();
		close(s);
		return;
	}

	gtk_threads_enter();
	conn_task_update(sess, 1);
	gtk_threads_leave();
#ifdef USE_IPV6

	if (connect(s, he->ai_addr, he->ai_addrlen)) {
#else
	if(connect(s, (struct sockaddr *)&saddr, sizeof(saddr))) {
#endif
			gtk_threads_enter();
			hx_printf_prefix(htlc, 0, INFOPREFIX, "connect: %s\n",

							 strerror(errno));
			setbtns(sess, 0);
			set_status_bar(0);
			set_disconnect_btn(sess, 0);
			conn_task_update(sess, 2);
			gtk_threads_leave();
			close(s);
			return;
	}


	gtk_threads_enter();
	set_status_bar(1);
	hx_printf_prefix(htlc, 0, INFOPREFIX, "%s %s\n", _("connected to"),

					 server_addr);
	gtk_threads_leave();

	{

		char magic[HTLS_MAGIC_LEN];
		fd_set rfds;
		struct timeval tv;

		write(s, HTLC_MAGIC, HTLC_MAGIC_LEN);

		FD_ZERO(&rfds);
		FD_SET(s, &rfds);
		tv.tv_sec = 30;
		tv.tv_usec = 0;
		select(s+1, &rfds, NULL, NULL, &tv);


		if(FD_ISSET(s, &rfds)) {
			read(s, magic, HTLS_MAGIC_LEN);
			if(strncmp(HTLS_MAGIC, magic, HTLS_MAGIC_LEN)) {
				gtk_threads_enter();
				hx_printf_prefix(htlc, 0, INFOPREFIX,
								 _("invalid hotline server\n"));
				setbtns(sess, 0);
				set_status_bar(0);
				set_disconnect_btn(sess, 0);
				conn_task_update(sess, 2);
				gtk_threads_leave();
				close(s);
				return;
			}
		}
		else {
			gtk_threads_enter();
			hx_printf_prefix(htlc, 0, INFOPREFIX,

							 _("no response from server after thirty seconds"));
			setbtns(sess, 0);
			set_status_bar(0);
			set_disconnect_btn(sess, 0);
			conn_task_update(sess, 2);
			gtk_threads_leave();
			close(s);
			return;
		}
	}
#ifdef USE_IPV6
	htlc->addr = he;
#else
	htlc->addr = saddr;
#endif

	htlc->fd = s;
	htlc->trans = 1;

	memset(&htlc->in, 0, sizeof(struct qbuf));
	memset(&htlc->out, 0, sizeof(struct qbuf));

	htlc->rcv = hx_rcv_hdr;
	qbuf_set(&htlc->in, 0, SIZEOF_HL_HDR);

	gtk_threads_enter();
	conn_task_update(sess, 2);
	gtk_threads_leave();

	fd_blocking(s, 0);
	fd_closeonexec(s, 1);

	hxd_files[s].ready_read = htlc_read;
	hxd_files[s].ready_write = htlc_write;
	hxd_files[s].conn.htlc = htlc;


	connected = 1;
	hxd_fd_add(s);
	htlc->gdk_input = 1;

	hxd_fd_set(s, FDR);

	if (login)
		strcpy(htlc->login, login);

	gtk_threads_enter();
	task_new(htlc, rcv_task_login, 0, 0, "login");
	gtk_threads_leave();


	icon16 = htons(htlc->icon);
	if (login) {
		llen = strlen(login);
		if (llen > 64)
			llen = 64;
		hl_encode(enclogin, login, llen);
	} else
		llen = 0;

	gtk_threads_enter();

	if (pass) {
		plen = strlen(pass);
		if (plen > 64)
			plen = 64;
		hl_encode(encpass, pass, plen);
		hlwrite(htlc, HTLC_HDR_LOGIN, 0, 4,
				HTLC_DATA_ICON, 2, &icon16,
				HTLC_DATA_LOGIN, llen, enclogin,
				HTLC_DATA_PASSWORD, plen, encpass,
				HTLC_DATA_NAME, strlen(htlc->name), htlc->name);
	}
	else {
		hlwrite(htlc, HTLC_HDR_LOGIN, 0, 3,
				HTLC_DATA_ICON, 2, &icon16,
				HTLC_DATA_LOGIN, llen, enclogin,
				HTLC_DATA_NAME, strlen(htlc->name), htlc->name);
	}
	gtk_threads_leave();

	g_free(cdata->login);
	g_free(cdata->pass);
	g_free(cdata->serverstr);
	g_free(cdata);

	conn_tid = 0;
#ifndef USE_DEBUG
	pthread_exit(0);
#endif
}


void hx_connect(struct htlc_conn *htlc, const char *serverstr, guint16 port,

				const char *login, const char *pass)
{
#ifndef USE_DEBUG
	pthread_attr_t attr;
#endif
	struct connect_data *cdata;

#ifndef USE_DEBUG
	if(conn_tid) {
		pthread_cancel(conn_tid);
		conn_tid = 0;
	}

	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
#endif


	if (htlc->fd) {
		hx_htlc_close(htlc);
	}
	hx_clear_chat(htlc, 0, 1);

	cdata = g_malloc(sizeof(struct connect_data));
	cdata->htlc = htlc;
	cdata->serverstr = g_strdup(serverstr);
	cdata->port = port;
	cdata->login = g_strdup(login);
	cdata->pass = g_strdup(pass);

#ifdef USE_DEBUG
	hx_thread_connect(cdata);
#else
	pthread_create(&conn_tid, &attr, (void *)&hx_thread_connect, (void *)cdata);
#endif
}

int htxf_connect (struct htxf_conn *htxf)
{
	struct htxf_hdr h;
	int s;

#ifdef USE_IPV6
	s = socket(htxf->listen_addr->ai_family, SOCK_STREAM, IPPROTO_TCP);
#else
	s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
#endif
	if (s < 0) {
		return -1;
	}

#ifdef USE_IPV6
	if (connect(s, htxf->listen_addr->ai_addr, htxf->listen_addr->ai_addrlen)){
#else
	if (connect(s, (struct sockaddr *)&htxf->listen_addr, sizeof(struct sockaddr))) {
#endif
		close(s);
		return -1;
	}

	h.magic = htonl(HTXF_MAGIC_INT);
	h.ref = htonl(htxf->ref);
	h.unknown = 0;
	h.len = htonl(htxf->total_size);
	if (write(s, &h, SIZEOF_HTXF_HDR) != SIZEOF_HTXF_HDR) {
		return -1;
	}

	return s;
}

static int b_read (int fd, void *bufp, size_t len)
{
	register guint8 *buf = (guint8 *)bufp;
	register int r, pos = 0;

	while (len) {
		if ((r = read(fd, &(buf[pos]), len)) <= 0)
			return -1;
		pos += r;
		len -= r;
	}

	return pos;
}

void hx_tracker_list(session *sess, char *serverstr, guint16 port)
{
	int s;
	guint16 nusers, nservers;
	unsigned char buf[HOSTLEN];
	unsigned char name[512], desc[512];
	struct in_addr a;
	int total;
	int i;
#ifdef USE_IPV6
	struct addrinfo *he;
	struct addrinfo hints;
	int error;
	char portstr[HOSTLEN];

	memset(&hints, 0, sizeof(struct addrinfo));
	hints.ai_family = PF_UNSPEC;
	hints.ai_flags = AI_CANONNAME;
	hints.ai_socktype = SOCK_STREAM;

	sprintf(portstr, "%u", port);
#else
	struct sockaddr_in saddr;

	memset(&saddr, 0, sizeof(struct sockaddr_in));
	saddr.sin_port = htons(port);
	saddr.sin_family = AF_INET;
#endif

#ifndef USE_IPV6
	if(!inet_pton(AF_INET, serverstr, &saddr.sin_addr)) {
		struct hostent *he;


		if((he = gethostbyname(serverstr))) {
			size_t len = (unsigned)he->h_length > sizeof(struct in_addr)
				? sizeof(struct in_addr) : he->h_length;
			memcpy(&saddr.sin_addr, he->h_addr, len);
		}
		else {
#else
			if((error = getaddrinfo(serverstr, portstr, &hints, &he))) {
#endif
				gtk_threads_enter();
#ifdef USE_IPV6
				hx_printf_prefix(&sessions[0].htlc, 0, INFOPREFIX,

								 "%s: %s\n", serverstr, gai_strerror(error));
#else
# ifdef HAVE_HSTRERROR
				hx_printf_prefix(&sessions[0].htlc, 0, INFOPREFIX,
								 "DNS lookup for %s failed: %s\n",
								 serverstr, hstrerror(h_errno));
# else
				hx_printf_prefix(&sessions[0].htlc, 0, INFOPREFIX,
								 "DNS lookup for %s failed\n", serverstr);
# endif
#endif
				gtk_threads_leave();
				return;
			}
#ifndef USE_IPV6
		}
#endif


#ifdef USE_IPV6
	if ((s = socket(he->ai_family, SOCK_STREAM, IPPROTO_TCP)) < 0) {
#else
	if ((s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
#endif
		gtk_threads_enter();
		hx_printf_prefix(&sessions[0].htlc, 0, INFOPREFIX, "tracker: %s: %s",

						 serverstr, strerror(errno));
		return;
	}


	fd_blocking(s, 1);


#ifdef USE_IPV6
	if((connect(s, he->ai_addr, he->ai_addrlen)) < 0) {
#else
	if((connect(s, (struct sockaddr *)&saddr, sizeof(struct sockaddr)))) {
#endif
		gtk_threads_enter();
		hx_printf_prefix(&sessions[0].htlc, 0, INFOPREFIX, "tracker: %s: %s",

						 serverstr, strerror(errno));
		gtk_threads_leave();
		return;
	}


	if (write(s, HTRK_MAGIC, HTRK_MAGIC_LEN) != HTRK_MAGIC_LEN)
		goto funk_dat;


	if (b_read(s, buf, 14) != 14) {
		goto funk_dat;
	}


	nservers = ntohs(*((guint16 *)(&(buf[10]))));
	total = nservers;
	for (i = 1; nservers; nservers--, i++) {
		if (b_read(s, buf, 8) == -1) {
			break;
		}


		if (!buf[0]) {	/* assuming an address does not begin with 0,

						   we can skip this */
			nservers++;
			continue;
		}


		if (b_read(s, buf+8, 3) == -1) {
			break;
		}



		a.s_addr = *((guint32 *)buf);
		port = ntohs(*((guint16 *)(&(buf[4]))));
		nusers = ntohs(*((guint16 *)(&(buf[6]))));


		if (b_read(s, name, (size_t)buf[10]) == -1) {
			break;
		}


		name[(size_t)buf[10]] = 0;
		CR2LF(name, (size_t)buf[10]);
		strip_ansi(name, (size_t)buf[10]);


		if (b_read(s, buf, 1) == -1) {
			break;
		}


		memset(desc, 0, sizeof(desc));


		if (b_read(s, desc, (size_t)buf[0]) == -1) {
			break;
		}


		desc[(size_t)buf[0]] = 0;
		CR2LF(desc, (size_t)buf[0]);
		strip_ansi(desc, (size_t)buf[0]);


		gtk_threads_enter();
		hx_output.tracker_server_create(a, port, nusers, name, desc, total);
		track_prog_update(sess, serverstr, i, total);
		gtk_threads_leave();
	}
  funk_dat:


	close(s);


	return;
}

void kill_threads(void) {
	pthread_cancel(conn_tid);
}

void hlwrite (struct htlc_conn *htlc, guint32 type, guint32 flag, int hc, ...)
{
	va_list ap;
	struct hl_hdr h;
	struct hl_data_hdr dhs;
	struct qbuf *q;
	guint32 this_off, pos;


	if (!htlc->fd) {
		return;
	}


	q = &htlc->out;
	this_off = q->pos + q->len;
	pos = this_off + SIZEOF_HL_HDR;
	q->len += SIZEOF_HL_HDR;
	q->buf = g_realloc(q->buf, q->pos + q->len);


	h.type = htonl(type);
	h.trans = htonl(htlc->trans);
	htlc->trans++;


/*	h.uid = htons(htlc->current_uid);
	h.uid = 0; */
	h.flag = htonl(flag);
	h.hc = htons(hc);


	va_start (ap, hc);
	while (hc) {
		guint16 t, l;
		guint8 *data;


		t = (guint16)va_arg (ap, int);
		l = (guint16)va_arg (ap, int);
		dhs.type = htons (t);
		dhs.len = htons (l);


		q->len += SIZEOF_HL_DATA_HDR + l;
		q->buf = g_realloc (q->buf, q->pos + q->len);
		memcpy (&q->buf[pos], (guint8 *)&dhs, 4);
		pos += 4;
		data = va_arg (ap, guint8 *);
		if (l) {
			memcpy (&q->buf[pos], data, l);
			pos += l;
		}
		hc--;
	}
	va_end(ap);


	h.len = h.len2= htonl((pos - this_off) - (SIZEOF_HL_HDR - sizeof(h.hc)));


	memcpy(q->buf + this_off, &h, SIZEOF_HL_HDR);


	hxd_fd_set(htlc->fd, FDW);
}

void hl_code (void *__dst, const void *__src, size_t len)
{
	guint8 *dst = (guint8 *)__dst, *src = (guint8 *)__src;

	for (; len; len--)
		*dst++ = ~*src++;
}
