/*
 * Copyright (C) 2001 Misha Nasledov <misha@nasledov.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <sys/time.h>
#include <time.h>
#include <netinet/in.h>
#include "hotline.h"
#include "hxd.h"
#include "hx.h"
#include "network.h"
#include "gtkutil.h"
#include "gtkhx.h"
#include "gtk_hlist.h"
#include "tasks.h"
#include "rcv.h"
#include "files.h"


void hx_news15_get_post(struct htlc_conn *htlc, struct news_item *item)
{
	guint8 *hldir;
	guint16 hldirlen;
	guint16 postid;

	hldir = path_to_hldir(item->group->path, &hldirlen, 0);
	task_new(htlc, rcv_task_news_post, item, 0, "news_post");


	postid = htons(item->postid);

	hlwrite(htlc, HTLC_HDR_GETTHREAD, 0, 3, HTLC_DATA_NEWSPATH, hldirlen, hldir, HTLC_DATA_THREADID, 2, &postid, HTLC_DATA_NEWSTYPE, strlen(item->parts[0].mime_type), item->parts[0].mime_type);
}

void hx_news15_cat_list(struct htlc_conn *htlc, char *path)
{
	guint8 *hldir;
	guint16 hldirlen;
	char *cpath;


	cpath = g_malloc(strlen(path));
	strcpy(cpath, path);
	hldir = path_to_hldir(path, &hldirlen, 0);
	task_new(htlc, rcv_task_newscat_list, cpath, 0, "news_category");
	hlwrite(htlc, HTLC_HDR_NEWSCATLIST, 0, 1, HTLC_DATA_NEWSPATH, hldirlen, hldir);
}

void hx_news15_fldr_list(struct htlc_conn *htlc, char *path)
{
	guint8 *hldir;

	guint16 hldirlen;
	char *pathn;

	pathn = g_malloc(strlen(path));
	strcpy(pathn, path);
	hldir = path_to_hldir(path, &hldirlen, 0);
	task_new(htlc, rcv_task_newsfolder_list, pathn, 0, "news_folder");

	hlwrite(htlc, HTLC_HDR_NEWSDIRLIST, 0, 1, HTLC_DATA_NEWSPATH, hldirlen, hldir);
}


struct gnews_folder {
	GtkWidget *window;
	GtkWidget *news_list;
	gint row, col;
	struct news_folder *news;
	struct gnews_folder *next, *prev;
};

struct gnews_folder *gfnews_list = 0;

extern void hx_news15_fldr_list(struct htlc_conn *htlc, char *path);
extern void hx_news15_cat_list(struct htlc_conn *htlc, char *path);

struct gnews_folder *gfnews_with_hlist (GtkWidget *hlist);


static void newsf_clicked(GtkWidget *widget, GdkEventButton *event)
{
	struct gnews_folder *gfnews;
	int row, col;


	gfnews = gfnews_with_hlist(widget);
	if(!gfnews)
		return;

	gtk_hlist_get_selection_info(GTK_HLIST(widget),
								 event->x, event->y, &row, &col);


	if(event->type == GDK_2BUTTON_PRESS) {
		struct folder_item *item = gtk_hlist_get_row_data(GTK_HLIST(widget), gfnews->row);
		if(item) {
			char path[4096];


			sprintf(path, "%s/%s", gfnews->news->path, item->name);
			if(item->type == 1) {


				hx_news15_fldr_list(&sessions[0].htlc, path);
			}
			else {
				hx_news15_cat_list(&sessions[0].htlc, path);
			}
		}
	}
	else {
		gfnews->row = row;
		gfnews->col = col;
	}
}

struct gnews_folder *gfnews_with_hlist (GtkWidget *hlist)
{
	struct gnews_folder *gfnews;

	for(gfnews = gfnews_list; gfnews; gfnews = gfnews->prev) {
		if(gfnews->news_list == hlist) {
			return gfnews;
		}
	}
	return 0;
}

struct gnews_folder *gfnews_with_path (char *path)
{
	struct gnews_folder *gfnews;

	for(gfnews = gfnews_list; gfnews; gfnews = gfnews->prev) {
		if(!strcmp(gfnews->news->path, path)) {
			return gfnews;
		}
	}
	return 0;
}

void delete_gfnews(struct gnews_folder *gfnews)
{
	if (gfnews->next)

		gfnews->next->prev = gfnews->prev;
	if (gfnews->prev)
		gfnews->prev->next = gfnews->next;
	if (gfnews == gfnews_list)
		gfnews_list = gfnews->prev;
	g_free(gfnews);
}

void destroy_gfnews_browser(GtkWidget *widget, gpointer data)
{
	struct gnews_folder *gfnews = gtk_object_get_data(GTK_OBJECT(widget), "gfnews");

	delete_gfnews(gfnews);
	gtk_widget_destroy(widget);
}

struct gnews_folder *create_gfnews_window(struct news_folder *news)
{
	struct gnews_folder *gfnews = g_malloc(sizeof(struct gnews_folder));
	GtkWidget *news_window;
	GtkWidget *news_list;
	GtkWidget *news_scroll;
	GtkWidget *topframe;
	GtkWidget *hbuttonbox;
	GtkWidget *vbox;
	GtkStyle *style;


	gfnews->prev = 0;
	gfnews->next = 0;
	gfnews->news = news;

	if(gfnews_list) {
		gfnews_list->next = gfnews;
		gfnews->prev = gfnews_list;
	}

	news_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_policy(GTK_WINDOW(news_window), 1, 1, 0);

	gtk_widget_realize(news_window);
	style = gtk_widget_get_style(news_window);
	gtk_widget_set_usize(news_window, 264, 400);
	gtk_window_set_title(GTK_WINDOW(news_window), news->path);
	gtk_object_set_data(GTK_OBJECT(news_window), "gfnews", gfnews);
	gtk_signal_connect(GTK_OBJECT(news_window), "delete_event", GTK_SIGNAL_FUNC(destroy_gfnews_browser), 0);

	news_scroll = gtk_scrolled_window_new(0, 0);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(news_scroll), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

	news_list = gtk_hlist_new(1);
	gtk_hlist_set_column_width(GTK_HLIST(news_list), 0, 64);
	gtk_hlist_set_column_width(GTK_HLIST(news_list), 1, 240);
	gtk_hlist_set_row_height(GTK_HLIST(news_list), 18);
	gtk_hlist_set_shadow_type(GTK_HLIST(news_list), GTK_SHADOW_NONE);
	gtk_hlist_set_column_justification(GTK_HLIST(news_list), 0, GTK_JUSTIFY_LEFT);
	gtk_signal_connect(GTK_OBJECT(news_list), "button_press_event", GTK_SIGNAL_FUNC(newsf_clicked), 0);

	topframe = gtk_frame_new(0);
	gtk_widget_set_usize(topframe, -2, 30);
	gtk_frame_set_shadow_type(GTK_FRAME(topframe), GTK_SHADOW_OUT);

	hbuttonbox = gtk_hbox_new(0,0);

	vbox = gtk_vbox_new(0, 0);
	gtk_widget_set_usize(vbox, 240, 400);
	gtk_container_add(GTK_CONTAINER(topframe), hbuttonbox);
	gtk_box_pack_start(GTK_BOX(vbox), topframe, 0, 0, 0);
	gtk_container_add(GTK_CONTAINER(news_scroll), news_list);
	gtk_box_pack_start(GTK_BOX(vbox), news_scroll, 1, 1, 0);
	gtk_container_add(GTK_CONTAINER(news_window), vbox);

	gfnews->window = news_window;
	gfnews->news_list = news_list;
	gtk_widget_show_all(news_window);
	init_keyaccel(news_window);
	gfnews_list = gfnews;

	return gfnews;
}

void output_news_folder (struct news_folder *news)
{
	struct gnews_folder *gfnews;
	GtkWidget *news_list;
	gint row, i;

	gfnews = gfnews_with_path(news->path);
	if(!gfnews) {
		gfnews = create_gfnews_window(news);
	}
	news_list = gfnews->news_list;


	gtk_hlist_clear(GTK_HLIST(news_list));
	gtk_hlist_freeze(GTK_HLIST(news_list));

	for(i = 0; i < news->num_entries; i++) {
		struct folder_item *item = news->entry[i];
		gchar *name[1] = {item->name};


		row = gtk_hlist_append(GTK_HLIST(news_list), name);
		gtk_hlist_set_row_data(GTK_HLIST(news_list), row, item);
	}
	gtk_hlist_thaw(GTK_HLIST(news_list));
}

struct gnews_catalog {
	struct news_group *group;
	struct gnews_catalog *next, *prev;
	GtkWidget *window;
	GtkWidget *news_tree;
	GtkWidget *news_text;
	GtkCTreeNode *row, *col;
};

struct gnews_catalog *gcnews_list;

struct gnews_catalog *gcnews_with_path(char *path)
{
	struct gnews_catalog *gcnews;

	for(gcnews = gcnews_list; gcnews; gcnews = gcnews->prev) {
		if(!strcmp(path, gcnews->group->path)) {
			return gcnews;
		}
	}
	return 0;
}

struct gnews_catalog *gcnews_with_group (struct news_group *group)
{
	struct gnews_catalog *gcnews;

	for(gcnews = gcnews_list; gcnews; gcnews = gcnews->prev) {
		if(group == gcnews->group) {
			return gcnews;
		}
	}

	return 0;
}

void delete_gcnews(struct gnews_catalog *gcnews)
{
	if (gcnews->next)

		gcnews->next->prev = gcnews->prev;
	if (gcnews->prev)
		gcnews->prev->next = gcnews->next;
	if (gcnews == gcnews_list)
		gcnews_list = gcnews->prev;


	g_free(gcnews);
}


static void destroy_gcnews_browser(GtkWidget *widget, gpointer data)
{
	struct gnews_catalog *gcnews = gtk_object_get_data(GTK_OBJECT(widget), "gcnews");

	delete_gcnews(gcnews);
	gtk_widget_destroy(widget);
}


extern void hx_news15_get_post(struct htlc_conn *htlc, struct news_item *item);
void newsc_clicked (GtkCTree *ctree, GList *node, gint column, gpointer data)

{
	struct news_item *item = gtk_ctree_node_get_row_data(ctree, (GtkCTreeNode *)node);

	hx_news15_get_post(&sessions[0].htlc, item);
}

struct gnews_catalog *create_gcnews_window(struct news_group *group)
{
	struct gnews_catalog *gcnews = g_malloc(sizeof(struct gnews_catalog));
	GtkWidget *news_window;
	GtkWidget *news_tree;
	GtkWidget *news_scroll;
	GtkWidget *topframe;
	GtkWidget *hbuttonbox;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *news_text;
	GtkStyle *style;


	gcnews->prev = gcnews_list;
	gcnews->next = 0;
	gcnews->group = group;

	if(gcnews_list) {
		gcnews_list->next = gcnews;
	}

	news_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_policy(GTK_WINDOW(news_window), 1, 1, 0);

	gtk_widget_realize(news_window);
	style = gtk_widget_get_style(news_window);
//	gtk_widget_set_usize(news_window, 264, 400);
	gtk_window_set_title(GTK_WINDOW(news_window), group->path);
	gtk_object_set_data(GTK_OBJECT(news_window), "gcnews", gcnews);
	gtk_signal_connect(GTK_OBJECT(news_window), "delete_event", GTK_SIGNAL_FUNC(destroy_gcnews_browser), 0);

	news_scroll = gtk_scrolled_window_new(0, 0);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(news_scroll), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

	news_tree = gtk_ctree_new(1, 0);
/*	gtk_clist_set_row_height(GTK_CLIST(news_tree), 18); */
	gtk_clist_set_shadow_type(GTK_CLIST(news_tree), GTK_SHADOW_NONE);
	gtk_signal_connect(GTK_OBJECT(news_tree), "tree_select_row", GTK_SIGNAL_FUNC(newsc_clicked), 0);

	topframe = gtk_frame_new(0);
	gtk_widget_set_usize(topframe, -2, 30);
	gtk_frame_set_shadow_type(GTK_FRAME(topframe), GTK_SHADOW_OUT);

	hbuttonbox = gtk_hbox_new(0,0);

	vbox = gtk_vbox_new(0, 0);
	gtk_widget_set_usize(vbox, 240, 400);
	gtk_container_add(GTK_CONTAINER(topframe), hbuttonbox);
	gtk_box_pack_start(GTK_BOX(vbox), topframe, 0, 0, 0);
	gtk_container_add(GTK_CONTAINER(news_scroll), news_tree);
	gtk_box_pack_start(GTK_BOX(vbox), news_scroll, 1, 1, 0);

	hbox = gtk_hbox_new(0, 0);
	gtk_box_pack_start(GTK_BOX(hbox), vbox, 0, 0, 0);

	gtk_container_add(GTK_CONTAINER(news_window), hbox);
	news_text = gtk_text_new(0, 0);
	gtk_box_pack_start(GTK_BOX(hbox), news_text, 0, 0, 0);

	gcnews->window = news_window;
	gcnews->news_tree = news_tree;
	gcnews->news_text = news_text;
	gtk_widget_show_all(news_window);
	init_keyaccel(news_window);
	gcnews_list = gcnews;

	return gcnews;
}

void output_news_catalog(struct news_group *group)
{
	struct gnews_catalog *gcnews;
	int i;
	GtkCTreeNode *parent;

	gcnews = gcnews_with_path(group->path);
	if(!gcnews) {
		gcnews = create_gcnews_window(group);
	}
	if(!gcnews) {
		return;
	}


	gtk_clist_clear(GTK_CLIST(gcnews->news_tree));
	gtk_clist_freeze(GTK_CLIST(gcnews->news_tree));
	for(i = 0; i < group->post_count; i++) {
		struct news_item *item = &(group->posts[i]);
		int j;
		gchar *text[1] = {item->subject};

		parent = 0;


		for(j = 0; j < group->post_count; j++) {
			if(group->posts[j].postid == item->parentid) {
				parent = group->posts[j].node;
				break;
			}
		}


		item->node = gtk_ctree_insert_node(GTK_CTREE(gcnews->news_tree), parent, 0, text, 0, 0, 0, 0, 0, 0, 0);
		gtk_ctree_node_set_row_data(GTK_CTREE(gcnews->news_tree), item->node, item);
	}
	gtk_clist_thaw(GTK_CLIST(gcnews->news_tree));
}

void output_news_thread(struct news_post *post)
{
	struct gnews_catalog *gcnews = gcnews_with_group(post->item->group);

	if(!gcnews) {
		return;
	}


	gtk_editable_delete_text(GTK_EDITABLE(gcnews->news_text), 0, -1);
	gtk_text_insert(GTK_TEXT(gcnews->news_text), 0, 0, 0, post->buf, strlen(post->buf));
}
