/*
 * Copyright 2001 Misha Nasledov <misha@nasledov.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <gtk/gtk.h>
#include <sys/time.h>
#include <time.h>
#include <netinet/in.h>
#include "hotline.h"
#include "hxd.h"
#include "hx.h"
#include "gtk_hlist.h"
#include "gtkhx.h"
#include "news.h"
#include "xtext.h"
#include "regex.h"
#include "cicn.h"
#include "sound.h"
#include "users.h"
#include "dfa.h"

time_t start_time;
time_t total_time;

static struct icon_viewer *iv;

extern struct msgwin *msg_list;
struct gtkhx_prefs gtkhx_prefs;
static void parse_tracker (void);
void parse_icons (void);

struct icon_viewer {
	guint32 icon_high;
	unsigned int nfound;
	GtkWidget *icon_list;
};

void list_icons (void)
{
	GdkPixmap *pixmap;
	GdkBitmap *mask;
	gint row;
	gchar *text[2] = {0, 0};
	guint32 icon;
	unsigned int nfound = 0;
	char buf[16];
	GtkWidget *icon_list = iv->icon_list;
	char *nam = "";
	guint16 nres = 0;
	int i;
	guint32 id;
	GdkColormap *colormap = gtk_widget_get_colormap(icon_list);
	GdkVisual *visual = gtk_widget_get_visual(icon_list);
	GdkGC *gc = NULL;
	GdkImage *image;
	GdkImage *maskim;
	int off;

	text[1] = buf;
	gtk_hlist_freeze(GTK_HLIST(icon_list));
	for (i = 0; i < icon_files.n; ++i) {
		if (!icon_files.cicns[i]) continue;
		nres = macres_file_num_res_of_type(icon_files.cicns[i], TYPE_cicn);
		for(icon = 0; icon < nres; icon++) {
			macres_res *r = 0;
			r = macres_file_get_nth_res_of_type(icon_files.cicns[i], TYPE_cicn, icon);
			if(!r) continue;

			image = (GdkImage *)cicn_to_gdkimage(colormap, visual, r->data, r->datalen, &maskim);

			off = image->width > 400 ? 198 : 0;
			pixmap = gdk_pixmap_new(icon_list->window, image->width-off, image->height, image->depth);
			if (!gc) gc = gdk_gc_new (pixmap);
			gdk_draw_image(pixmap, gc, image, off, 0, 0, 0, image->width-off, image->height);
			if (maskim) {
				mask = gdk_pixmap_new(icon_list->window, image->width-off, image->height, 1);
				if (!mask_gc) mask_gc = gdk_gc_new (mask);
				gdk_draw_image(mask, mask_gc, maskim, off, 0, 0, 0, image->width-off, image->height);
				gdk_image_destroy(maskim);
			}
			else
				mask = 0;

			gdk_image_destroy(image);
			if(!pixmap) continue;

			nfound++;
			sprintf(buf, "%u", r->resid);
			row = gtk_hlist_append(GTK_HLIST(icon_list), text);
			id = r->resid;
			gtk_hlist_set_row_data(GTK_HLIST(icon_list), row, (gpointer)id);
			gtk_hlist_set_pixtext(GTK_HLIST(icon_list), row, 0, nam, 34, pixmap, mask);
			gdk_pixmap_unref(pixmap);
			if(mask) gdk_pixmap_unref(mask);

/*			g_free (r->name); */
			g_free (r);

			/* Cooperative multitasking */
			if (i % 10 == 0) {
				while (gtk_events_pending()) gtk_main_iteration();

				if (!gtkhx_prefs.gtk.options_window)/*Have we been destroyed?!*/
					return;
			}
		}
	}
	if(nfound >= 2) {
		gtk_hlist_sort(GTK_HLIST(icon_list));
	}
	gtk_hlist_thaw(GTK_HLIST(icon_list));
}

static void
icon_row_selected (GtkWidget *widget, gint row, gint column, GdkEventButton *event, gpointer data)
{
	guint16 icon;
	char buf[16];

	if(!GTK_HLIST(widget)->rows) return;
	icon = GPOINTER_TO_INT(gtk_hlist_get_row_data(GTK_HLIST(widget), row));
	if(!icon) {
		return;
	}
	sprintf(buf, "%u", icon);
	gtk_entry_set_text(GTK_ENTRY(gtkhx_prefs.gtk.icon), buf);
}

void reinit_gtktexts (session *sess)
{
	struct msgwin *msg;
	struct gtkhx_chat *gchat;

	if(gtktext_style) {
		gtk_style_unref(gtktext_style);
	}
	gtktext_style = gtk_style_new();

	gtktext_style->fg[GTK_STATE_NORMAL] = fg_col;
	gtktext_style->text[GTK_STATE_NORMAL] = fg_col;
	gtktext_style->light[GTK_STATE_NORMAL] = fg_col;

	gtktext_style->bg[GTK_STATE_NORMAL] = bg_col;
	gtktext_style->base[GTK_STATE_NORMAL] = bg_col;
	gtktext_style->dark[GTK_STATE_NORMAL] = bg_col;

	gtktext_style->font = gtkhx_font;

	if(gtkhx_prefs.geo.news.open) {
		gtk_widget_set_style(sess->news_text, gtktext_style);
	}
	for(gchat = sess->gchat_list; gchat; gchat = gchat->prev) {
		gtk_xtext_set_font(GTK_XTEXT(gchat->output), gtkhx_font, 0);
		gtk_xtext_refresh(GTK_XTEXT(gchat->output), 1);
		if(gchat->input) {
			gtk_widget_set_style(gchat->input, gtktext_style);
		}
		if(gchat->subject) {
			gtk_widget_set_style(gchat->subject, gtktext_style);
		}
	}
	for(msg = sess->msg_list; msg; msg = msg->prev) {
		gtk_xtext_set_font(GTK_XTEXT(msg->outputbuf), gtkhx_font, 0);
		gtk_xtext_refresh(GTK_XTEXT(gchat->output), 1);
		gtk_widget_set_style(msg->inputbuf, gtktext_style);
	}
}

void init_variables() /* default settings if prefs file is not found. */
{
	gtktext_style = gtk_style_new();

	gtkhx_prefs.geo.chat.xsize = 412;
	gtkhx_prefs.geo.chat.ysize = 312;
	gtkhx_prefs.geo.chat.xpos = 0;
	gtkhx_prefs.geo.chat.ypos = 0;
	gtkhx_prefs.geo.chat.open = 0;
	gtkhx_prefs.geo.chat.init = 1;

	gtkhx_prefs.geo.news.xsize = 412;
	gtkhx_prefs.geo.news.ysize = 384;
	gtkhx_prefs.geo.news.xpos = 0;
	gtkhx_prefs.geo.news.ypos = 0;
	gtkhx_prefs.geo.news.open = 0;
	gtkhx_prefs.geo.news.init = 1;

	gtkhx_prefs.geo.users.xsize = 264;
	gtkhx_prefs.geo.users.ysize = 400;
	gtkhx_prefs.geo.users.xpos = 0;
	gtkhx_prefs.geo.users.ypos = 0;
	gtkhx_prefs.geo.users.open = 0;
	gtkhx_prefs.geo.users.init = 1;

	gtkhx_prefs.geo.tool.xpos = 0;
	gtkhx_prefs.geo.tool.ypos = 0;

	gtkhx_prefs.geo.tasks.xsize = 264;
	gtkhx_prefs.geo.tasks.ysize = 100;
	gtkhx_prefs.geo.tasks.xpos = 0;
	gtkhx_prefs.geo.tasks.ypos = 0;
	gtkhx_prefs.geo.tasks.open = 0;
	gtkhx_prefs.geo.tasks.init = 1;

	gtkhx_prefs.icon_str = g_strdup_printf("%s/share/gtkhx/icons.rsrc,%s/share/gtkhx/bficons.rsrc", PREFIX, PREFIX);
	gtkhx_prefs.sound_path = g_strdup_printf("%s/share/gtkhx/sounds/", PREFIX);
	gtkhx_prefs.download_path = g_strdup(".");

	hxsnd.invite = 1;
	hxsnd.chat = 1;
	hxsnd.error = 1;
	hxsnd.file = 1;
	hxsnd.join = 1;
	hxsnd.login = 1;
	hxsnd.msg = 1;
	hxsnd.news = 1;
	hxsnd.part = 1;
	hxsnd.on = 0;

	gtkhx_prefs.queuedl = 1;
	gtkhx_prefs.showback = 0;
	gtkhx_prefs.showjoin = 1;
	gtkhx_prefs.auto_reply = 0;
	gtkhx_prefs.auto_reply_msg = 0;
	gtkhx_prefs.trans_xtext = 0;
#ifdef USE_GDK_PIXBUF
	gtkhx_prefs.tint_red = 255;
	gtkhx_prefs.tint_green = 255;
	gtkhx_prefs.tint_blue = 255;
#endif
	gtkhx_prefs.use_fontset = 0;
	gtkhx_prefs.xbuf_max = 500;

	gtkhx_font = gdk_font_load("-b&h-lucidatypewriter-medium-r-normal-*-*-100-*-*-m-*-iso8859-1");

	if(!gtkhx_font) {
		gtkhx_prefs.font = "fixed";
		gtkhx_font = gdk_font_load("fixed"); // fall back to the fixed font if ! default font
	}
	else {
		gtkhx_prefs.font = g_strdup("-b&h-lucidatypewriter-medium-r-normal-*-*-100-*-*-m-*-iso8859-1");
	}

	fg_col.red = 0xcccc;
	fg_col.green = 0xcccc;
	fg_col.blue = 0xcccc;
	bg_col.red = 0x0000;
	bg_col.green = 0x0000;
	bg_col.blue = 0x0000;

	gtktext_style->fg[GTK_STATE_NORMAL] = fg_col;
	gtktext_style->text[GTK_STATE_NORMAL] = fg_col;
	gtktext_style->light[GTK_STATE_NORMAL] = fg_col;

	gtktext_style->bg[GTK_STATE_NORMAL] = bg_col;
	gtktext_style->base[GTK_STATE_NORMAL] = bg_col;
	gtktext_style->dark[GTK_STATE_NORMAL] = bg_col;

	gtktext_style->font = gtkhx_font;

	gtkhx_prefs.timestamp = 0;
	gtkhx_prefs.word_wrap = 0;
	gtkhx_prefs.file_samewin = 1;
	gtkhx_prefs.track_case = 0;
	gtkhx_prefs.old_nickcompletion = 0;

	dfasyntax ((RE_CHAR_CLASSES | RE_CONTEXT_INDEP_ANCHORS |
				RE_CONTEXT_INDEP_OPS | RE_HAT_LISTS_NOT_NEWLINE |
				RE_NEWLINE_ALT | RE_NO_BK_PARENS | RE_NO_BK_VBAR),
			   !gtkhx_prefs.track_case, '\n');

	gtkhx_prefs.snd_cmd = g_strdup("play");

	gtkhx_prefs.tracker_str = g_strdup("hltracker.com");
	gtkhx_prefs.tracker = NULL;

	parse_tracker();
	parse_icons();

	start_time = time(NULL);
}


void prefs_allocate(char *tag, char *line) /* XXX: Redo prefs code someday */
{
	if(!strcmp(tag, "OLD_NICKCOMPLETION")) {
		gtkhx_prefs.old_nickcompletion = atoi(line);
	}
	if(strcmp(tag, "TRACKER") == 0) {
		if(gtkhx_prefs.tracker_str) {
			g_free(gtkhx_prefs.tracker_str);
		}
		gtkhx_prefs.tracker_str = g_strdup(line);
		parse_tracker();
 	}
	else if(strcmp(tag, "NICK") == 0) {
		hx_change_name_icon(&sessions[0].htlc, line, 0);
	}
	else if(strcmp(tag, "ICON") == 0) {
		hx_change_name_icon(&sessions[0].htlc, 0, atoi(line));
	}
	else if(strcmp(tag, "SHOWJOIN") == 0) {
		gtkhx_prefs.showjoin = atoi(line);
	}
	else if(strcmp(tag, "SHOWBACK") == 0) {
		gtkhx_prefs.showback = atoi(line);
	}
	else if(strcmp(tag, "QUEUEDL") == 0) {
		gtkhx_prefs.queuedl = atoi(line);
	}
	else if(strcmp(tag, "CHATXSIZE") == 0) {
		gtkhx_prefs.geo.chat.xsize = atoi(line);
	}
	else if(strcmp(tag, "CHATYSIZE") == 0) {
		gtkhx_prefs.geo.chat.ysize = atoi(line);
	}
	else if(strcmp(tag, "CHATXPOS") == 0) {
		gtkhx_prefs.geo.chat.xpos = atoi(line);
	}
	else if(strcmp(tag, "CHATYPOS") == 0) {
		gtkhx_prefs.geo.chat.ypos = atoi(line);
	}
	else if(strcmp(tag, "NEWSXSIZE") == 0) {
		gtkhx_prefs.geo.news.xsize = atoi(line);
	}
	else if(strcmp(tag, "NEWSYSIZE") == 0) {
		gtkhx_prefs.geo.news.ysize = atoi(line);
	}
	else if(strcmp(tag, "NEWSXPOS") == 0) {
		gtkhx_prefs.geo.news.xpos = atoi(line);
	}
	else if(strcmp(tag, "NEWSYPOS") == 0) {
		gtkhx_prefs.geo.news.ypos = atoi(line);
	}
	else if(strcmp(tag, "TOOLXPOS") == 0) {
		gtkhx_prefs.geo.tool.xpos = atoi(line);
	}
	else if(strcmp(tag, "TOOLYPOS") == 0) {
		gtkhx_prefs.geo.tool.ypos = atoi(line);
	}
	else if(strcmp(tag, "USERXSIZE") == 0) {
		gtkhx_prefs.geo.users.xsize = atoi(line);
	}
	else if(strcmp(tag, "USERYSIZE") == 0) {
		gtkhx_prefs.geo.users.ysize = atoi(line);
	}
	else if(strcmp(tag, "USERXPOS") == 0) {
		gtkhx_prefs.geo.users.xpos = atoi(line);
	}
	else if(strcmp(tag, "USERYPOS") == 0) {
		gtkhx_prefs.geo.users.ypos = atoi(line);
	}
	else if(strcmp(tag, "TASKXSIZE") == 0) {
		gtkhx_prefs.geo.tasks.xsize = atoi(line);
	}
	else if(strcmp(tag, "TASKYSIZE") == 0) {
		gtkhx_prefs.geo.tasks.ysize = atoi(line);
	}
	else if(strcmp(tag, "TASKXPOS") == 0) {
		gtkhx_prefs.geo.tasks.xpos = atoi(line);
	}
	else if(strcmp(tag, "TASKYPOS") == 0) {
		gtkhx_prefs.geo.tasks.ypos = atoi(line);
	}
	else if(strcmp(tag, "OPENCHAT") == 0) {
		gtkhx_prefs.geo.chat.init = atoi(line);
	}
	else if(strcmp(tag, "OPENNEWS") == 0) {
		gtkhx_prefs.geo.news.init = atoi(line);
	}
	else if(strcmp(tag, "OPENUSERS") == 0) {
		gtkhx_prefs.geo.users.init = atoi(line);
	}
	else if(strcmp(tag, "OPENTASKS") == 0) {
		gtkhx_prefs.geo.tasks.init = atoi(line);
	}
	else if(strcmp(tag, "ICONS") == 0) {
		if(gtkhx_prefs.icon_str) {
			g_free(gtkhx_prefs.icon_str);
		}
		gtkhx_prefs.icon_str = g_strdup(line);
		parse_icons();
	}
	else if(strcmp(tag, "SOUNDPATH") == 0) {
		if(gtkhx_prefs.sound_path) {
			g_free(gtkhx_prefs.sound_path);
		}
		gtkhx_prefs.sound_path = g_strdup(line);
	}
	else if(strcmp(tag, "SOUNDSON") == 0) {
		hxsnd.on = atoi(line);
	}
	else if(strcmp(tag, "SOUNDINVITE") == 0) {
		hxsnd.invite = atoi(line);
	}
	else if(strcmp(tag, "SOUNDCHAT") == 0) {
		hxsnd.chat = atoi(line);
	}
	else if(strcmp(tag, "SOUNDERROR") == 0) {
		hxsnd.error = atoi(line);
	}
	else if(strcmp(tag, "SOUNDFILE") == 0) {
		hxsnd.file = atoi(line);
	}
	else if(strcmp(tag, "SOUNDJOIN") == 0) {
		hxsnd.join = atoi(line);
	}
	else if(strcmp(tag, "SOUNDLOGIN") == 0) {
		hxsnd.login = atoi(line);
	}
	else if(strcmp(tag, "SOUNDMSG") == 0) {
		hxsnd.msg = atoi(line);
	}
	else if(strcmp(tag, "SOUNDNEWS") == 0) {
		hxsnd.news = atoi(line);
	}
	else if(strcmp(tag, "SOUNDPART") == 0) {
		hxsnd.part = atoi(line);
	}
	else if(strcmp(tag, "AUTOREPLYON") == 0) {
		gtkhx_prefs.auto_reply = atoi(line);
	}
	else if(strcmp(tag, "AUTOREPLYMSG") == 0) {
		if(gtkhx_prefs.auto_reply_msg) {
			g_free(gtkhx_prefs.auto_reply_msg);
		}
		gtkhx_prefs.auto_reply_msg = g_strdup(line);
	}
	else if(strcmp(tag, "TRANSPARENT") == 0) {
		gtkhx_prefs.trans_xtext = atoi(line);
	}
	else if(strcmp(tag, "FONTSET") == 0) {
		gtkhx_prefs.use_fontset = atoi(line);
	}
	else if(strcmp(tag, "FONT") == 0) {
		if(gtkhx_prefs.font) {
			g_free(gtkhx_prefs.font);
		}
		gtkhx_prefs.font = g_strdup(line);
		if(gtkhx_prefs.use_fontset) {
			gtkhx_font = gdk_fontset_load(gtkhx_prefs.font);
		}
		else {
			gtkhx_font = gdk_font_load(gtkhx_prefs.font);
		}
		if(!gtkhx_font) {
			gtkhx_font = gdk_font_load("fixed");
			if(gtkhx_prefs.font) {
				g_free(gtkhx_prefs.font);
			}
			gtkhx_prefs.font = g_strdup("fixed");
		}
	}
	else if(strcmp(tag, "DOWNLOAD") == 0) {
		if(gtkhx_prefs.download_path) {
			g_free(gtkhx_prefs.download_path);
		}
		gtkhx_prefs.download_path = g_strdup(line);
	}
#ifdef USE_GDK_PIXBUF
	else if(strcmp(tag, "RED") == 0) {
		gtkhx_prefs.tint_red = atoi(line);
	}
	else if(strcmp(tag, "GREEN") == 0) {
		gtkhx_prefs.tint_green = atoi(line);
	}
	else if(strcmp(tag, "BLUE") == 0) {
		gtkhx_prefs.tint_blue = atoi(line);
	}
#endif
	else if(!strcmp(tag, "TIMESTAMP")) {
		gtkhx_prefs.timestamp = atoi(line);
	}
	else if(!strcmp(tag, "SND_CMD")) {
		if(gtkhx_prefs.snd_cmd) {
			g_free(gtkhx_prefs.snd_cmd);
		}
		gtkhx_prefs.snd_cmd = g_strdup(line);
	}
	else if(!strcmp(tag, "WORDWRAP")) {
		gtkhx_prefs.word_wrap = atoi(line);
	}
	else if(!strcmp(tag, "FILE_SAMEWINDOW")) {
		gtkhx_prefs.file_samewin = atoi(line);
	}
	else if(!strcmp(tag, "TRACKER_CASE")) {
		gtkhx_prefs.track_case = atoi(line);
		dfasyntax ((RE_CHAR_CLASSES | RE_CONTEXT_INDEP_ANCHORS |
					RE_CONTEXT_INDEP_OPS | RE_HAT_LISTS_NOT_NEWLINE |
					RE_NEWLINE_ALT | RE_NO_BK_PARENS | RE_NO_BK_VBAR),
				   !gtkhx_prefs.track_case, '\n');
	}
	else if(!strcmp(tag, "XBUF_MAX")) {
		gtkhx_prefs.xbuf_max = atoi(line);
	}
	else if(!strcmp(tag, "TIME")) {
		total_time = atoi(line);
	}
}

void parse_line(char *line)
{
	char *result, *tag;
	char *rsltptr;
	size_t lenr, lenl = strlen(line);

	rsltptr = strchr(line, '=');
	lenr = strlen(rsltptr);
	result = g_malloc(lenr+1);
	strcpy(result, rsltptr);
	rsltptr = result; rsltptr++;
	result[strlen(result)-1] = 0;
	tag = g_malloc(1+(lenl-lenr));
	strncpy(tag, line, (lenl-lenr));
	tag[(lenl-lenr)] = 0;

	prefs_allocate(tag, rsltptr);
	g_free(tag);
	g_free(result);
}

char *read_line(FILE *prefs)
{
	char *line = NULL;
	size_t size = 0, pos = 0;

	do {
		size_t len;
		size += 128;
		line = g_realloc(line, size);
		if (!(fgets(line+pos, size-pos, prefs) && (len=strlen(line+pos)))) {
			g_free(line);
			return NULL;
		}
		pos += len;
	} while (line[pos-1] != '\n');
	return line;
}

void prefs_read(void)
{
#ifndef WIN32
	char *home = getenv("HOME");
	char *path = g_strdup_printf("%s/.gtkhxrc", home);
#else
	char *path = g_strdup("gtkhxrc");
#endif
	char *prefsline;
	FILE *prefs = fopen(path, "r");


	if(!prefs) {
		g_warning(_("Cannot open %s for reading\n"), path);
		return;
	}

	while((prefsline = read_line(prefs))) {
		if(*prefsline != '#' && *prefsline != '\n') {
			parse_line(prefsline);
		}
		g_free(prefsline);
	}

	fclose(prefs);
	g_free(path);
}

void prefs_write(void)
{
#ifndef WIN32
	char *home = getenv("HOME");
	char *path = g_strdup_printf("%s/.gtkhxrc", home);
#else
	char *path = g_strdup("gtkhxrc");
#endif
 	FILE *prefs = fopen(path, "w");
	time_t now;

	if(!prefs) {
		g_warning(_("Could not write to %s/.gtkhxrc\n"), home);
		return;
	}

	now = time(NULL);
	total_time += (now-start_time);
	start_time = now;

	fprintf(prefs, "# This is the GtkHx preferences file\n");
	fprintf(prefs, "# Lines starting with '#' are comments\n\n");

	fprintf(prefs, "# General Settings\n");
	fprintf(prefs, "NICK=%s\n", sessions[0].htlc.name);
	fprintf(prefs, "ICON=%u\n", sessions[0].htlc.icon);
	fprintf(prefs, "TRACKER=%s\n", gtkhx_prefs.tracker_str);
	fprintf(prefs, "SHOWJOIN=%d\n", gtkhx_prefs.showjoin);
	fprintf(prefs, "SHOWBACK=%d\n", gtkhx_prefs.showback);
	fprintf(prefs, "QUEUEDL=%d\n", gtkhx_prefs.queuedl);
	fprintf(prefs, "FONTSET=%d\n", gtkhx_prefs.use_fontset);
	fprintf(prefs, "FONT=%s\n", gtkhx_prefs.font);
	fprintf(prefs, "ICONS=%s\n", gtkhx_prefs.icon_str);
	fprintf(prefs, "SOUNDPATH=%s\n", gtkhx_prefs.sound_path);
	fprintf(prefs, "DOWNLOAD=%s\n", gtkhx_prefs.download_path);
	fprintf(prefs, "FILE_SAMEWINDOW=%d\n", gtkhx_prefs.file_samewin);
	fprintf(prefs, "TRACKER_CASE=%d\n", gtkhx_prefs.track_case);
	fprintf(prefs, "OLD_NICKCOMPLETION=%d\n", gtkhx_prefs.old_nickcompletion);
	fprintf(prefs, "TIME=%lu\n", total_time);

	fprintf(prefs, "\n# XText Settings\n");
	fprintf(prefs, "TRANSPARENT=%d\n", gtkhx_prefs.trans_xtext);
#ifdef USE_GDK_PIXBUF
	fprintf(prefs, "RED=%d\n", gtkhx_prefs.tint_red);
	fprintf(prefs, "GREEN=%d\n", gtkhx_prefs.tint_green);
	fprintf(prefs, "BLUE=%d\n", gtkhx_prefs.tint_blue);
#endif
	fprintf(prefs, "TIMESTAMP=%d\n", gtkhx_prefs.timestamp);
 	fprintf(prefs, "WORDWRAP=%d\n", gtkhx_prefs.word_wrap);
	fprintf(prefs, "XBUF_MAX=%d\n", gtkhx_prefs.xbuf_max);

	fprintf(prefs, "\n# Chat Geometry\n");
	fprintf(prefs, "CHATXSIZE=%d\n", gtkhx_prefs.geo.chat.xsize);
	fprintf(prefs, "CHATYSIZE=%d\n", gtkhx_prefs.geo.chat.ysize);
	fprintf(prefs, "CHATXPOS=%d\n", gtkhx_prefs.geo.chat.xpos);
	fprintf(prefs, "CHATYPOS=%d\n", gtkhx_prefs.geo.chat.ypos);
	fprintf(prefs, "OPENCHAT=%d\n", gtkhx_prefs.geo.chat.open);

	fprintf(prefs, "\n# News Geometry\n");
	fprintf(prefs, "NEWSXSIZE=%d\n", gtkhx_prefs.geo.news.xsize);
	fprintf(prefs, "NEWSYSIZE=%d\n", gtkhx_prefs.geo.news.ysize);
	fprintf(prefs, "NEWSXPOS=%d\n", gtkhx_prefs.geo.news.xpos);
	fprintf(prefs, "NEWSYPOS=%d\n", gtkhx_prefs.geo.news.ypos);
	fprintf(prefs, "OPENNEWS=%d\n", gtkhx_prefs.geo.news.open);

	fprintf(prefs, "\n# Users Geometry\n");
	fprintf(prefs, "USERXSIZE=%d\n", gtkhx_prefs.geo.users.xsize);
	fprintf(prefs, "USERYSIZE=%d\n", gtkhx_prefs.geo.users.ysize);
	fprintf(prefs, "USERXPOS=%d\n", gtkhx_prefs.geo.users.xpos);
	fprintf(prefs, "USERYPOS=%d\n", gtkhx_prefs.geo.users.ypos);
	fprintf(prefs, "OPENUSERS=%d\n", gtkhx_prefs.geo.users.open);

	fprintf(prefs, "\n# Toolbar Geometry\n");
	fprintf(prefs, "TOOLXPOS=%d\n", gtkhx_prefs.geo.tool.xpos);
	fprintf(prefs, "TOOLYPOS=%d\n", gtkhx_prefs.geo.tool.ypos);

	fprintf(prefs, "\n# Tasks Geometry\n");
	fprintf(prefs, "TASKXSIZE=%d\n", gtkhx_prefs.geo.tasks.xsize);
	fprintf(prefs, "TASKYSIZE=%d\n", gtkhx_prefs.geo.tasks.ysize);
	fprintf(prefs, "TASKXPOS=%d\n", gtkhx_prefs.geo.tasks.xpos);
	fprintf(prefs, "TASKYPOS=%d\n", gtkhx_prefs.geo.tasks.ypos);
	fprintf(prefs, "OPENTASKS=%d\n", gtkhx_prefs.geo.tasks.open);

	fprintf(prefs, "\n# Sounds\n");
	fprintf(prefs, "SND_CMD=%s\n", gtkhx_prefs.snd_cmd);
	fprintf(prefs, "SOUNDSON=%d\n", hxsnd.on);
	fprintf(prefs, "SOUNDINVITE=%d\n", hxsnd.invite);
	fprintf(prefs, "SOUNDCHAT=%d\n", hxsnd.chat);
	fprintf(prefs, "SOUNDERROR=%d\n", hxsnd.error);
	fprintf(prefs, "SOUNDFILE=%d\n", hxsnd.file);
	fprintf(prefs, "SOUNDJOIN=%d\n", hxsnd.join);
	fprintf(prefs, "SOUNDLOGIN=%d\n", hxsnd.login);
	fprintf(prefs, "SOUNDMSG=%d\n", hxsnd.msg);
	fprintf(prefs, "SOUNDNEWS=%d\n", hxsnd.news);
	fprintf(prefs, "SOUNDPART=%d\n", hxsnd.part);

	fprintf(prefs, "\n# Auto Reply\n");
	fprintf(prefs, "AUTOREPLYON=%d\n", gtkhx_prefs.auto_reply);
	fprintf(prefs, "AUTOREPLYMSG=%s\n", gtkhx_prefs.auto_reply_msg);

	fclose(prefs);
	g_free(path);
}

void parse_icons (void)
{
	char *com, *icons = gtkhx_prefs.icon_str;
	int i;

	if(gtkhx_prefs.icon) {
		for (i = 0; i != gtkhx_prefs.num_icons; ++i) {
			g_free(gtkhx_prefs.icon[i]);
		}
		g_free(gtkhx_prefs.icon);
		gtkhx_prefs.icon = NULL;
	}
	gtkhx_prefs.num_icons = 0;
	if(!*icons || !*(icons+1)) return;
	for (i=0; ; ++i) {
		if (!(com = strchr (icons, ','))) com = &icons[strlen(icons)];
		gtkhx_prefs.num_icons++;
		gtkhx_prefs.icon = g_realloc(gtkhx_prefs.icon, (i+1)*sizeof(char*));
		gtkhx_prefs.icon[i] = g_malloc(com-icons+1);
		memcpy(gtkhx_prefs.icon[i], icons, com-icons);
		gtkhx_prefs.icon[i][com-icons] = '\0';
		if (!*com) break;
		icons = com+1;
	}
}

static void parse_tracker (void)
{
	char *com, *trackers = gtkhx_prefs.tracker_str;
	int i;

	if(gtkhx_prefs.tracker) {
		for (i = 0; i != gtkhx_prefs.num_tracker; ++i) {
			g_free(gtkhx_prefs.tracker[i]);
		}
		g_free(gtkhx_prefs.tracker);
		gtkhx_prefs.tracker = NULL;
	}
	gtkhx_prefs.num_tracker = 0;
	if(!*trackers || !*(trackers+1)) return;
	for (i=0; ; ++i) {
		if (!(com = strchr (trackers, ','))) com = &trackers[strlen(trackers)];
		gtkhx_prefs.num_tracker++;
		gtkhx_prefs.tracker = g_realloc(gtkhx_prefs.tracker, (i+1)*sizeof(char*));
		gtkhx_prefs.tracker[i] = g_malloc(com-trackers+1);
		memcpy(gtkhx_prefs.tracker[i], trackers, com-trackers);
		gtkhx_prefs.tracker[i][com-trackers] = '\0';
		if (!*com) break;
		trackers = com+1;
	}
}

static void parse_tracker_list(void)
{
	GtkWidget *list = gtkhx_prefs.gtk.tracker_list;
	int i = 0;
	size_t len = 0;

	/* go through the CList and populate the gtkhx_prefs.tracker
	   array and create gtkhx_prefs.tracker_str */
	if(gtkhx_prefs.tracker) {
		for (i = 0; i != gtkhx_prefs.num_tracker; ++i) {
			g_free(gtkhx_prefs.tracker[i]);
		}
	}
	g_free(gtkhx_prefs.tracker);

	gtkhx_prefs.num_tracker = GTK_CLIST(list)->rows;
	gtkhx_prefs.tracker = g_malloc(GTK_CLIST(list)->rows * sizeof(char*));
	g_free (gtkhx_prefs.tracker_str); gtkhx_prefs.tracker_str = g_malloc0(1);

	for(i = 0; i < GTK_CLIST(list)->rows; i++) {
		char *tracker = gtk_clist_get_row_data(GTK_CLIST(list), i);
		size_t trackersize = strlen(tracker)+1;
		gtkhx_prefs.tracker_str = g_realloc(gtkhx_prefs.tracker_str, len+trackersize+1);
		if (i) {
			gtkhx_prefs.tracker_str[len] = ',';
			memcpy(gtkhx_prefs.tracker_str+len+1, tracker, trackersize);
			len++;
		}
		else
			memcpy(gtkhx_prefs.tracker_str+len, tracker, trackersize);
		len += trackersize-1;
		gtkhx_prefs.tracker[i] = g_strdup(tracker);
	}
}

void options_change (GtkWidget *widget, gpointer data)
{
	char *nicknam;
	char *iconstr;
	guint16 icon;
	int fsetl_tmp;
	struct gtkhx_chat *gchat;
	struct msgwin *msg;

	session *sess = gtk_object_get_data(GTK_OBJECT(gtkhx_prefs.gtk.options_window), "sess");

	iconstr = gtk_entry_get_text(GTK_ENTRY(gtkhx_prefs.gtk.icon));
	nicknam = gtk_entry_get_text(GTK_ENTRY(gtkhx_prefs.gtk.nick));

	if(gtkhx_prefs.sound_path) {
		g_free(gtkhx_prefs.sound_path);
	}
	gtkhx_prefs.sound_path = g_strdup(gtk_entry_get_text(GTK_ENTRY(gtkhx_prefs.gtk.sound_path)));

	if(gtkhx_prefs.icon_str) {
		g_free(gtkhx_prefs.icon_str);
	}
	gtkhx_prefs.icon_str = g_strdup(gtk_entry_get_text(GTK_ENTRY(gtkhx_prefs.gtk.icon_path)));

	gtkhx_prefs.showjoin = gtk_toggle_button_get_active((GtkToggleButton *)gtkhx_prefs.gtk.showjoin);
	gtkhx_prefs.showback = gtk_toggle_button_get_active((GtkToggleButton *)gtkhx_prefs.gtk.showback);
	gtkhx_prefs.old_nickcompletion = gtk_toggle_button_get_active((GtkToggleButton *)gtkhx_prefs.gtk.old_nickcompletion);

	if(gtkhx_prefs.download_path) {
		g_free(gtkhx_prefs.download_path);
	}
	gtkhx_prefs.download_path = g_strdup(gtk_entry_get_text(GTK_ENTRY(gtkhx_prefs.gtk.downloadpath)));


	if(gtkhx_prefs.snd_cmd) {
		g_free(gtkhx_prefs.snd_cmd);
	}
	gtkhx_prefs.snd_cmd = g_strdup(gtk_entry_get_text(GTK_ENTRY(gtkhx_prefs.gtk.snd_cmd)));

	if(!*gtkhx_prefs.download_path) {
		g_free(gtkhx_prefs.download_path);
		gtkhx_prefs.download_path = g_strdup(".");
	}

	gtkhx_prefs.queuedl = gtk_toggle_button_get_active((GtkToggleButton *)gtkhx_prefs.gtk.queuedl);

	hxsnd.on = gtk_toggle_button_get_active((GtkToggleButton *)hxsnd.gtk.on);
	hxsnd.invite = gtk_toggle_button_get_active((GtkToggleButton *)hxsnd.gtk.inv);
	hxsnd.chat = gtk_toggle_button_get_active((GtkToggleButton *)hxsnd.gtk.chat);
	hxsnd.error = gtk_toggle_button_get_active((GtkToggleButton *)hxsnd.gtk.err);
	hxsnd.file = gtk_toggle_button_get_active((GtkToggleButton *)hxsnd.gtk.file);
	hxsnd.join = gtk_toggle_button_get_active((GtkToggleButton *)hxsnd.gtk.join);
	hxsnd.login = gtk_toggle_button_get_active((GtkToggleButton *)hxsnd.gtk.login);
	hxsnd.msg = gtk_toggle_button_get_active((GtkToggleButton *)hxsnd.gtk.msg);
	hxsnd.news = gtk_toggle_button_get_active((GtkToggleButton *)hxsnd.gtk.news);
	hxsnd.part = gtk_toggle_button_get_active((GtkToggleButton *)hxsnd.gtk.part);

	gtkhx_prefs.auto_reply = gtk_toggle_button_get_active((GtkToggleButton *)gtkhx_prefs.gtk.auto_reply);

	if(gtkhx_prefs.auto_reply_msg) {
		g_free(gtkhx_prefs.auto_reply_msg);
	}
	gtkhx_prefs.auto_reply_msg = g_strdup(gtk_entry_get_text(GTK_ENTRY(gtkhx_prefs.gtk.auto_reply_msg)));

	gtkhx_prefs.trans_xtext = gtk_toggle_button_get_active((GtkToggleButton *)gtkhx_prefs.gtk.trans_xtext);
	fsetl_tmp = gtk_toggle_button_get_active((GtkToggleButton *)gtkhx_prefs.gtk.use_fontset);

	gtkhx_prefs.timestamp = gtk_toggle_button_get_active((GtkToggleButton *)gtkhx_prefs.gtk.timestamp);
	gtkhx_prefs.word_wrap = gtk_toggle_button_get_active((GtkToggleButton *)gtkhx_prefs.gtk.word_wrap);
	gtkhx_prefs.file_samewin = gtk_toggle_button_get_active((GtkToggleButton *)gtkhx_prefs.gtk.file_samewin);
	gtkhx_prefs.track_case = gtk_toggle_button_get_active((GtkToggleButton *)gtkhx_prefs.gtk.track_case);

	gtkhx_prefs.xbuf_max = gtk_spin_button_get_value_as_int((GtkSpinButton *)gtkhx_prefs.gtk.xbuf_max);

	dfasyntax ((RE_CHAR_CLASSES | RE_CONTEXT_INDEP_ANCHORS |
				RE_CONTEXT_INDEP_OPS | RE_HAT_LISTS_NOT_NEWLINE |
				RE_NEWLINE_ALT | RE_NO_BK_PARENS | RE_NO_BK_VBAR),
			   !gtkhx_prefs.track_case, '\n');

	for(gchat = sess->gchat_list; gchat; gchat = gchat->prev) {
		gtk_xtext_set_background(GTK_XTEXT(gchat->output), 0,
								 gtkhx_prefs.trans_xtext, 1);
		GTK_XTEXT(gchat->output)->wordwrap = gtkhx_prefs.word_wrap;
		GTK_XTEXT(gchat->output)->max_lines = gtkhx_prefs.xbuf_max;
		gtk_xtext_refresh(GTK_XTEXT(gchat->output), 1);
	}
	for(msg = sess->msg_list; msg; msg = msg->prev) {
		gtk_xtext_set_background(GTK_XTEXT(msg->outputbuf), 0,
								 gtkhx_prefs.trans_xtext, 1);
		GTK_XTEXT(msg->outputbuf)->wordwrap = gtkhx_prefs.word_wrap;
		GTK_XTEXT(msg->outputbuf)->max_lines = gtkhx_prefs.xbuf_max;
		gtk_xtext_refresh(GTK_XTEXT(msg->outputbuf), 1);
	}


	if(!(strcmp(gtkhx_prefs.font, gtk_entry_get_text(GTK_ENTRY(gtkhx_prefs.gtk.font))) == 0) || gtkhx_prefs.use_fontset != fsetl_tmp) {
		gtkhx_prefs.use_fontset = fsetl_tmp;
		if(gtkhx_prefs.font) {
			g_free(gtkhx_prefs.font);
		}
		gtkhx_prefs.font = g_strdup(gtk_entry_get_text(GTK_ENTRY(gtkhx_prefs.gtk.font)));
		if(gtkhx_prefs.use_fontset) {
			gtkhx_font = gdk_fontset_load(gtkhx_prefs.font);
		}
		else {
			gtkhx_font = gdk_font_load(gtkhx_prefs.font);
		}
		if(!gtkhx_font) {
			g_free(gtkhx_prefs.font);
			gtkhx_prefs.font = g_strdup("fixed");
			gtkhx_font = gdk_font_load(gtkhx_prefs.font);
		}
		reinit_gtktexts(sess);
	}

	icon = strtoul(iconstr, 0, 0);
	hx_change_name_icon(&sess->htlc, nicknam, icon);

	parse_tracker_list();
	parse_icons();
	init_icons();
	prefs_write();

	if(!GPOINTER_TO_INT(data)) {
		gtk_widget_destroy(gtkhx_prefs.gtk.options_window);
		gtkhx_prefs.gtk.options_window = NULL;
		g_free(iv);
	}
}

static void
close_options_window (void)
{
	gtk_widget_destroy(gtkhx_prefs.gtk.options_window);
	gtkhx_prefs.gtk.options_window = 0;
	g_free(iv);
}

void set_font (GtkWidget *btn, GtkWidget *fontsel)
{
	GtkWidget *entry = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(btn), "entry");

	gtk_entry_set_text(GTK_ENTRY(entry), gtk_font_selection_dialog_get_font_name(GTK_FONT_SELECTION_DIALOG(fontsel)));

	gtk_widget_destroy(fontsel);
}

void create_fontsel (GtkWidget *btn, GtkWidget *entry)
{
	GtkWidget *fontsel = gtk_font_selection_dialog_new(_("Browse Fonts"));

	gtk_signal_connect_object(GTK_OBJECT(GTK_FONT_SELECTION_DIALOG(fontsel)->cancel_button), "clicked", (GtkSignalFunc) gtk_widget_destroy, GTK_OBJECT(fontsel));

	gtk_object_set_data(GTK_OBJECT(GTK_FONT_SELECTION_DIALOG(fontsel)->ok_button), "entry", entry);
	gtk_signal_connect(GTK_OBJECT(GTK_FONT_SELECTION_DIALOG(fontsel)->ok_button), "clicked", GTK_SIGNAL_FUNC(set_font), fontsel);
	gtk_font_selection_dialog_set_font_name(GTK_FONT_SELECTION_DIALOG(fontsel), gtkhx_prefs.font);

	gtk_widget_show_all(fontsel);
}

#ifdef USE_GDK_PIXBUF
static void
settings_slider_cb (GtkAdjustment * adj, int *value)
{
	struct gtkhx_chat *gchat;
	struct msgwin *msg;
	int i;

	*value = adj->value;
	for(i = 0; i < sizeof(sessions)/sizeof(sessions[0]); i++) {
		for(gchat = sessions[i].gchat_list; gchat; gchat = gchat->prev) {
			GTK_XTEXT (gchat->output)->tint_red = gtkhx_prefs.tint_red;
			GTK_XTEXT (gchat->output)->tint_green = gtkhx_prefs.tint_green;
			GTK_XTEXT (gchat->output)->tint_blue = gtkhx_prefs.tint_blue;
			gtk_xtext_refresh (GTK_XTEXT (gchat->output), gtkhx_prefs.trans_xtext);
		}
	}
	for(msg = msg_list; msg; msg = msg->prev) {
		GTK_XTEXT (msg->outputbuf)->tint_red = gtkhx_prefs.tint_red;
		GTK_XTEXT (msg->outputbuf)->tint_green = gtkhx_prefs.tint_green;
		GTK_XTEXT (msg->outputbuf)->tint_blue = gtkhx_prefs.tint_blue;
		gtk_xtext_refresh (GTK_XTEXT (msg->outputbuf), gtkhx_prefs.trans_xtext);
	}
}
static void
settings_slider (GtkWidget * vbox, char *label, int *value)
{
	GtkAdjustment *adj;
	GtkWidget *wid, *hbox, *lbox, *lab;

	hbox = gtk_hbox_new (0, 0);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, 0, 0, 2);

	lbox = gtk_hbox_new (0, 0);
	gtk_box_pack_start (GTK_BOX (hbox), lbox, 0, 0, 2);
	gtk_widget_set_usize (lbox, 60, 0);

	lab = gtk_label_new (label);
	gtk_box_pack_end (GTK_BOX (lbox), lab, 0, 0, 0);

	adj = (GtkAdjustment *) gtk_adjustment_new (*value, 0, 255.0, 1, 25, 0);
	gtk_signal_connect (GTK_OBJECT (adj), "value_changed",
							  GTK_SIGNAL_FUNC (settings_slider_cb), value);

	wid = gtk_hscale_new (adj);
	gtk_scale_set_value_pos ((GtkScale *) wid, GTK_POS_RIGHT);
	gtk_scale_set_digits ((GtkScale *) wid, 0);
	gtk_container_add (GTK_CONTAINER (hbox), wid);
}

#endif

static GtkWidget *
settings_create_group (GtkWidget * vvbox, gchar * title)
{
	GtkWidget *frame;
	GtkWidget *vbox;

	frame = gtk_frame_new (title);
	gtk_box_pack_start (GTK_BOX (vvbox), frame, FALSE, FALSE, 0);

	vbox = gtk_vbox_new (FALSE, 2);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 2);
	gtk_container_add (GTK_CONTAINER (frame), vbox);

	return vbox;
}

static void add_tracker(GtkWidget *add, GtkWidget *entry)
{
	char *tracker = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry)));
	gint row;
	gchar *text[] = {tracker};

	row = gtk_clist_append(GTK_CLIST(gtkhx_prefs.gtk.tracker_list), text);
	gtk_clist_set_row_data(GTK_CLIST(gtkhx_prefs.gtk.tracker_list), row, tracker);
	gtk_entry_set_text(GTK_ENTRY(entry), "");
}

static void remove_tracker(GtkWidget *del, GtkWidget *list)
{
	gtk_clist_remove(GTK_CLIST(list), GTK_CLIST(list)->focus_row);
}

static void settings_page_tracker (GtkWidget *vbox)
{
	GtkWidget *wid;
	GtkWidget *btnhbox, *ent_hbox;
	GtkWidget *add, *remove;
	GtkWidget *lbl, *entry;
	GtkWidget *scroll;
	int i, row;

	wid = settings_create_group(vbox, _("Trackers"));

	scroll = gtk_scrolled_window_new(0, 0);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
				       GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);

	gtkhx_prefs.gtk.tracker_list = gtk_clist_new(1);
	gtk_container_add(GTK_CONTAINER(scroll), gtkhx_prefs.gtk.tracker_list);

	gtk_box_pack_start(GTK_BOX(wid), scroll, 0, 0, 0);
	gtk_widget_set_usize(scroll, 232, 246);

	ent_hbox = gtk_hbox_new(0, 0);
	gtk_box_pack_start(GTK_BOX(wid), ent_hbox, 0, 0, 0);

	lbl = gtk_label_new(_("Address: "));
	entry = gtk_entry_new();

	gtk_box_pack_start(GTK_BOX(ent_hbox), lbl, 0, 0, 0);
	gtk_box_pack_start(GTK_BOX(ent_hbox), entry, 0, 0, 0);

	btnhbox = gtk_hbox_new(0, 0);
	gtk_box_pack_start(GTK_BOX(wid), btnhbox, 0, 0, 0);

	add = gtk_button_new_with_label(_("Add"));
	gtk_signal_connect(GTK_OBJECT(add), "clicked", GTK_SIGNAL_FUNC(add_tracker), entry);

	remove = gtk_button_new_with_label(_("Remove"));
	gtk_signal_connect(GTK_OBJECT(remove), "clicked", GTK_SIGNAL_FUNC(remove_tracker), gtkhx_prefs.gtk.tracker_list);

	gtk_box_pack_start(GTK_BOX(btnhbox), add, 0, 0, 0);
	gtk_box_pack_start(GTK_BOX(btnhbox), remove, 0, 0, 0);

	for(i = 0; i < gtkhx_prefs.num_tracker; i++) {
		gchar *text[] = {gtkhx_prefs.tracker[i]};
		char *tracker = g_strdup(gtkhx_prefs.tracker[i]);

		row = gtk_clist_append(GTK_CLIST(gtkhx_prefs.gtk.tracker_list), text);
		gtk_clist_set_row_data(GTK_CLIST(gtkhx_prefs.gtk.tracker_list), row, tracker);
	}
}



static void settings_page_files(GtkWidget *vbox)
{
	GtkWidget *wid;
	GtkWidget *table;

	wid = settings_create_group(vbox, _("File Browsing"));
	table = gtk_table_new(1, 1, 0);

	gtkhx_prefs.gtk.file_samewin = gtk_check_button_new_with_label(_("Browse in Same Window"));
	gtk_toggle_button_set_active((GtkToggleButton *)gtkhx_prefs.gtk.file_samewin, gtkhx_prefs.file_samewin);

	gtk_table_attach(GTK_TABLE(table), gtkhx_prefs.gtk.file_samewin, 0, 1, 0, 1, GTK_EXPAND|GTK_FILL, 0, 0, 0);

	gtk_box_pack_start(GTK_BOX(wid), table, 0, 0, 0);
}

static void settings_page_sound(GtkWidget *vbox)
{
	GtkWidget *wid;
	GtkWidget *table;
	GtkWidget *table2;
	GtkWidget *lbl;

	wid = settings_create_group(vbox, _("Sound Command"));
	table2 = gtk_table_new(1, 2, 0);

	table = gtk_table_new(10, 2, 0);
	gtk_table_set_col_spacings(GTK_TABLE(table2), 5);

	lbl = gtk_label_new(_("Sound Command: "));
	gtk_table_attach(GTK_TABLE(table2), lbl, 0, 1, 2, 3,
					 GTK_FILL, GTK_FILL, 0, 0);

	gtkhx_prefs.gtk.snd_cmd = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(gtkhx_prefs.gtk.snd_cmd), gtkhx_prefs.snd_cmd);
	gtk_table_attach(GTK_TABLE(table2), gtkhx_prefs.gtk.snd_cmd, 1, 2, 2, 3,
					 (GTK_EXPAND|GTK_FILL), 0, 0, 0);

	gtk_box_pack_start(GTK_BOX(wid), table2, 0, 0, 0);

	wid = settings_create_group(vbox, _("Sounds"));

	hxsnd.gtk.on = gtk_check_button_new_with_label(_("Play Sounds For:"));
	gtk_toggle_button_set_active((GtkToggleButton *)hxsnd.gtk.on, hxsnd.on);
	hxsnd.gtk.inv = gtk_check_button_new_with_label(_("Chat Invitation"));
	gtk_toggle_button_set_active((GtkToggleButton *)hxsnd.gtk.inv, hxsnd.invite);
	hxsnd.gtk.chat = gtk_check_button_new_with_label(_("Chat"));
	gtk_toggle_button_set_active((GtkToggleButton *)hxsnd.gtk.chat, hxsnd.chat);
	hxsnd.gtk.err = gtk_check_button_new_with_label(_("Error"));
	gtk_toggle_button_set_active((GtkToggleButton *)hxsnd.gtk.err, hxsnd.error);
	hxsnd.gtk.file = gtk_check_button_new_with_label(_("File Transfer Complete"));
	gtk_toggle_button_set_active((GtkToggleButton *)hxsnd.gtk.file, hxsnd.file);
	hxsnd.gtk.join = gtk_check_button_new_with_label(_("Join"));
	gtk_toggle_button_set_active((GtkToggleButton *)hxsnd.gtk.join, hxsnd.join);
	hxsnd.gtk.login = gtk_check_button_new_with_label(_("Login"));
	gtk_toggle_button_set_active((GtkToggleButton *)hxsnd.gtk.login, hxsnd.login);
	hxsnd.gtk.msg = gtk_check_button_new_with_label(_("Private Message"));
	gtk_toggle_button_set_active((GtkToggleButton *)hxsnd.gtk.msg, hxsnd.msg);
	hxsnd.gtk.news = gtk_check_button_new_with_label(_("News"));
	gtk_toggle_button_set_active((GtkToggleButton *)hxsnd.gtk.news, hxsnd.news);
	hxsnd.gtk.part = gtk_check_button_new_with_label(_("Leave"));
	gtk_toggle_button_set_active((GtkToggleButton *)hxsnd.gtk.part, hxsnd.part);

	gtk_table_attach(GTK_TABLE(table), hxsnd.gtk.on, 0, 1, 0, 1, GTK_EXPAND|GTK_FILL, 0, 0, 4);
	gtk_table_attach(GTK_TABLE(table), hxsnd.gtk.inv, 1, 2, 1, 2, GTK_EXPAND|GTK_FILL, 0, 0, 0);

	gtk_table_attach(GTK_TABLE(table), hxsnd.gtk.chat, 1, 2, 2, 3, GTK_EXPAND|GTK_FILL, 0, 0, 0);

	gtk_table_attach(GTK_TABLE(table), hxsnd.gtk.err, 1, 2, 3, 4, GTK_EXPAND|GTK_FILL, 0, 0, 0);

	gtk_table_attach(GTK_TABLE(table), hxsnd.gtk.file, 1, 2, 4, 5, GTK_EXPAND|GTK_FILL, 0, 0, 0);

	gtk_table_attach(GTK_TABLE(table), hxsnd.gtk.join, 1, 2, 5, 6, GTK_EXPAND|GTK_FILL, 0, 0, 0);

	gtk_table_attach(GTK_TABLE(table), hxsnd.gtk.login, 1, 2, 6, 7, GTK_EXPAND|GTK_FILL, 0, 0, 0);

	gtk_table_attach(GTK_TABLE(table), hxsnd.gtk.msg, 1, 2, 7, 8, GTK_EXPAND|GTK_FILL, 0, 0, 0);

	gtk_table_attach(GTK_TABLE(table), hxsnd.gtk.news, 1, 2, 8, 9, GTK_EXPAND|GTK_FILL, 0, 0, 0);

	gtk_table_attach(GTK_TABLE(table), hxsnd.gtk.part, 1, 2, 9, 10, GTK_EXPAND|GTK_FILL, 0, 0, 0);

	gtk_box_pack_start(GTK_BOX(wid), table, 0, 0, 0);
}

static void settings_page_font(GtkWidget *vbox)
{
	GtkWidget *wid;
	GtkWidget *table;
	GtkWidget *table2;
	GtkWidget *lbl;
	GtkWidget *btn;

	wid = settings_create_group(vbox, _("Fonts"));

	table = gtk_table_new(1, 2, 0);

	lbl = gtk_label_new(_("Font: "));
	gtk_table_attach(GTK_TABLE(table), lbl, 0, 1, 2, 3,
					 GTK_FILL, GTK_FILL, 0, 0);

	gtkhx_prefs.gtk.font = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(gtkhx_prefs.gtk.font), gtkhx_prefs.font);
	gtk_table_attach(GTK_TABLE(table), gtkhx_prefs.gtk.font, 1, 2, 2, 3,
					 (GTK_EXPAND|GTK_FILL), 0, 0, 0);

	gtk_box_pack_start(GTK_BOX(wid), table, 0, 0, 0);

	table2 = gtk_table_new(2, 1, 0);

	gtkhx_prefs.gtk.use_fontset = gtk_check_button_new_with_label(_("Use gdk_fontset_load"));
	gtk_toggle_button_set_active((GtkToggleButton *)gtkhx_prefs.gtk.use_fontset, gtkhx_prefs.use_fontset);
	gtk_table_attach(GTK_TABLE(table2), gtkhx_prefs.gtk.use_fontset, 0, 1, 3, 4, GTK_FILL, GTK_FILL, 0, 0);

	btn = gtk_button_new_with_label(_("Browse Fonts"));
	gtk_table_attach(GTK_TABLE(table2), btn, 0, 1, 4, 5, GTK_FILL, GTK_FILL, 0, 0);
	gtk_signal_connect(GTK_OBJECT(btn), "clicked", GTK_SIGNAL_FUNC(create_fontsel), gtkhx_prefs.gtk.font);

	gtk_box_pack_start(GTK_BOX(wid), table2, 0, 0, 0);
}

static void settings_page_xtext(GtkWidget *vbox)
{
	GtkWidget *lbl;
	GtkWidget *wid;
	GtkWidget *table;
	GtkWidget *table2;
	GtkAdjustment *adj;

	wid = settings_create_group(vbox, _("Transparency"));

	table = gtk_table_new(4, 1, 0);
	gtk_table_set_row_spacings(GTK_TABLE(table), 10);
	gtk_container_set_border_width(GTK_CONTAINER(table), 10);

	gtkhx_prefs.gtk.trans_xtext = gtk_check_button_new_with_label(_("Use Transparent XText Widget"));
	gtk_toggle_button_set_active((GtkToggleButton *) gtkhx_prefs.gtk.trans_xtext, gtkhx_prefs.trans_xtext);
	gtk_table_attach(GTK_TABLE(table), gtkhx_prefs.gtk.trans_xtext, 0, 1, 0, 1, GTK_FILL, 0, 0, 0);
	gtk_box_pack_start(GTK_BOX(wid), table, 0, 0, 0);

#ifdef USE_GDK_PIXBUF
	settings_slider(wid, _("Red: "), &gtkhx_prefs.tint_red);
	settings_slider(wid, _("Green: "), &gtkhx_prefs.tint_green);
	settings_slider(wid, _("Blue: "), &gtkhx_prefs.tint_blue);
#endif



	wid = settings_create_group(vbox, _("Miscellaeneous"));

	table2 = gtk_table_new(2, 2, 0);

	gtkhx_prefs.gtk.timestamp = gtk_check_button_new_with_label(_("Timestamp Chat"));
	gtk_toggle_button_set_active((GtkToggleButton *)gtkhx_prefs.gtk.timestamp, gtkhx_prefs.timestamp);
	gtk_table_attach(GTK_TABLE(table2), gtkhx_prefs.gtk.timestamp, 0, 1, 0, 1, GTK_FILL, 0, 0, 0);

	gtkhx_prefs.gtk.word_wrap = gtk_check_button_new_with_label(_("Word Wrap"));
	gtk_toggle_button_set_active((GtkToggleButton *)gtkhx_prefs.gtk.word_wrap, gtkhx_prefs.word_wrap);
	gtk_table_attach(GTK_TABLE(table2), gtkhx_prefs.gtk.word_wrap, 1, 2, 0, 1, GTK_FILL, 0, 0, 0);

	adj = (GtkAdjustment *)gtk_adjustment_new(gtkhx_prefs.xbuf_max, 0, 0xffff, 1, 1, 1);
	gtkhx_prefs.gtk.xbuf_max = gtk_spin_button_new(adj, 1, 0);
	lbl = gtk_label_new(_("Maximum Lines (0 = Unlimited)    "));

	gtk_table_attach(GTK_TABLE(table2), lbl, 0, 1, 1, 2, GTK_FILL, 0, 0, 0);
	gtk_table_attach(GTK_TABLE(table2), gtkhx_prefs.gtk.xbuf_max, 1, 2, 1, 2, GTK_FILL, 0, 0, 0);

	gtk_box_pack_start(GTK_BOX(wid), table2, 0, 0, 0);
}

static void settings_page_path(GtkWidget *vbox)
{
	GtkWidget *wid;
	GtkWidget *table;
	GtkWidget *lbl;
	GtkWidget *desc;

	wid = settings_create_group(vbox, _("Notes"));

	desc = gtk_label_new(_("You can have more then one icon list.\n"
						   "Enter as many icon lists as you want, comma separated."));

	gtk_box_pack_start(GTK_BOX(wid), desc, 0, 0, 0);

	wid = settings_create_group(vbox, _("Paths"));

	table = gtk_table_new(3, 2, 0);
	gtk_table_set_row_spacings(GTK_TABLE(table), 10);
	gtk_table_set_col_spacings(GTK_TABLE(table), 5);

	lbl = gtk_label_new(_("Icon Path:"));
	gtk_label_set_justify(GTK_LABEL(lbl), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment(GTK_MISC(lbl), 0, 0.5);
	gtk_table_attach(GTK_TABLE(table), lbl, 0, 1, 0, 1, GTK_FILL, (GtkAttachOptions) 0, 0, 0);

	gtkhx_prefs.gtk.icon_path = gtk_entry_new();
	gtk_table_attach(GTK_TABLE(table), gtkhx_prefs.gtk.icon_path, 1, 2, 0, 1,
					 (GtkAttachOptions) (GTK_EXPAND|GTK_FILL),
					 (GtkAttachOptions) 0, 0, 0);
	gtk_entry_set_text(GTK_ENTRY(gtkhx_prefs.gtk.icon_path), gtkhx_prefs.icon_str);

	lbl = gtk_label_new(_("Sound Path:"));
	gtk_label_set_justify(GTK_LABEL(lbl), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment(GTK_MISC(lbl), 0, 0.5);
	gtk_table_attach(GTK_TABLE(table), lbl, 0, 1, 1, 2, GTK_FILL, (GtkAttachOptions) 0, 0, 0);

	gtkhx_prefs.gtk.sound_path = gtk_entry_new();
	gtk_table_attach(GTK_TABLE(table), gtkhx_prefs.gtk.sound_path, 1, 2, 1, 2,
					 (GtkAttachOptions) (GTK_EXPAND|GTK_FILL),
					 (GtkAttachOptions) 0, 0, 0);
	gtk_entry_set_text(GTK_ENTRY(gtkhx_prefs.gtk.sound_path), gtkhx_prefs.sound_path);

	lbl = gtk_label_new(_("Download Path:"));
	gtk_label_set_justify(GTK_LABEL(lbl), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment(GTK_MISC(lbl), 0, 0.5);
	gtk_table_attach(GTK_TABLE(table), lbl, 0, 1, 2, 3, GTK_FILL, (GtkAttachOptions) 0, 0, 0);

	gtkhx_prefs.gtk.downloadpath = gtk_entry_new();
	gtk_table_attach(GTK_TABLE(table), gtkhx_prefs.gtk.downloadpath, 1, 2, 2, 3,
					 (GtkAttachOptions) (GTK_EXPAND|GTK_FILL),
					 (GtkAttachOptions) 0, 0, 0);
	gtk_entry_set_text(GTK_ENTRY(gtkhx_prefs.gtk.downloadpath), gtkhx_prefs.download_path);

	gtk_box_pack_start(GTK_BOX(wid), table, 0, 0, 0);
}

static int listsorthelper (GtkHList *hlist,
				GtkHListRow *ptr1,
				GtkHListRow *ptr2)
{
	int i1=atoi(GTK_HELL_TEXT(ptr1->cell[1])->text);
	int i2=atoi(GTK_HELL_TEXT(ptr2->cell[1])->text);
	if (i1 < i2) return -1;
	if (i1 > i2) return 1;
	return 0;
}

static void settings_page_icon(GtkWidget *vbox)
{
	GtkWidget *wid;
	GtkWidget *table;
	GtkWidget *scroll;
	GtkWidget *icon_list;
	GtkWidget *label;
	char iconstr[16];

	iv = g_malloc(sizeof(struct icon_viewer));

	wid = settings_create_group(vbox, _("Icon"));

	table = gtk_table_new(3, 2, 0);
	gtk_table_set_row_spacings(GTK_TABLE(table), 10);
	gtk_table_set_col_spacings(GTK_TABLE(table), 5);

	label = gtk_label_new(_("Icon ID: "));
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1, 0, 0, 0, 0);

	gtkhx_prefs.gtk.icon = gtk_entry_new();
	gtk_table_attach(GTK_TABLE(table), gtkhx_prefs.gtk.icon, 1, 2, 0, 1,
			 (GtkAttachOptions)(GTK_EXPAND | GTK_FILL),
			 (GtkAttachOptions)0, 0, 0);
	sprintf(iconstr, "%u", sessions[0].htlc.icon);
	gtk_entry_set_text(GTK_ENTRY(gtkhx_prefs.gtk.icon), iconstr);

	scroll = gtk_scrolled_window_new(0,0);

	icon_list = gtk_hlist_new(2);
	gtk_hlist_set_selection_mode(GTK_HLIST(icon_list), GTK_SELECTION_SINGLE);
	gtk_hlist_set_column_width(GTK_HLIST(icon_list), 0, 260);
	gtk_hlist_set_column_width(GTK_HLIST(icon_list), 1, 42);
	gtk_hlist_set_row_height(GTK_HLIST(icon_list), 18);
	gtk_hlist_set_compare_func(GTK_HLIST(icon_list), (GtkHListCompareFunc)listsorthelper);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
				       GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
	gtk_widget_set_usize(scroll, 232, 256);

	gtk_signal_connect(GTK_OBJECT(icon_list), "select_row",
			   GTK_SIGNAL_FUNC(icon_row_selected), iv);

	gtk_container_add(GTK_CONTAINER(scroll), icon_list);
	gtk_table_attach(GTK_TABLE(table), scroll, 0, 2, 1, 2,
			 (GtkAttachOptions)(GTK_EXPAND | GTK_FILL),
			 (GtkAttachOptions)0, 0, 0);

	iv->icon_list = icon_list;
	iv->nfound = 0;
	iv->icon_high = 0;

	gtk_box_pack_start(GTK_BOX(wid), table, 0, 0, 0);
}

static void settings_page_misc(GtkWidget *vbox)
{
	GtkWidget *wid;
	GtkWidget *table;

	wid = settings_create_group(vbox, _("Miscellaeneous"));

	table = gtk_table_new(7, 2, 0);
	gtk_table_set_row_spacings(GTK_TABLE(table), 10);
	gtk_table_set_col_spacings(GTK_TABLE(table), 5);


	gtkhx_prefs.gtk.auto_reply = gtk_check_button_new_with_label(_("Auto Reply"));
	gtk_toggle_button_set_active((GtkToggleButton *)gtkhx_prefs.gtk.auto_reply, gtkhx_prefs.auto_reply);
	gtk_table_attach(GTK_TABLE(table), gtkhx_prefs.gtk.auto_reply, 0, 1, 0, 1, GTK_FILL, 0, 0, 0);

	gtkhx_prefs.gtk.auto_reply_msg = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(gtkhx_prefs.gtk.auto_reply_msg), gtkhx_prefs.auto_reply_msg);
	gtk_table_attach(GTK_TABLE(table), gtkhx_prefs.gtk.auto_reply_msg, 1, 2, 0, 1, 0, 0, 0, 0);

	gtkhx_prefs.gtk.showback = gtk_check_button_new_with_label (_("Show Private Messages at Back"));
	gtk_toggle_button_set_active((GtkToggleButton *)gtkhx_prefs.gtk.showback, gtkhx_prefs.showback);
	gtk_table_attach(GTK_TABLE(table), gtkhx_prefs.gtk.showback, 0, 1, 1, 2,
			 (GtkAttachOptions)(GTK_FILL),
			 (GtkAttachOptions)(GTK_FILL), 0, 0);

	gtkhx_prefs.gtk.queuedl = gtk_check_button_new_with_label(_("Queue File Transfers"));
	gtk_toggle_button_set_active((GtkToggleButton *)gtkhx_prefs.gtk.queuedl, gtkhx_prefs.queuedl);
	gtk_table_attach(GTK_TABLE(table), gtkhx_prefs.gtk.queuedl, 0, 1, 2, 3,
			 (GtkAttachOptions)(GTK_FILL),
			 (GtkAttachOptions)(GTK_FILL), 0, 0);

	gtkhx_prefs.gtk.showjoin = gtk_check_button_new_with_label(_("Show Join/Leave in Chat"));
	gtk_toggle_button_set_active    ((GtkToggleButton *)gtkhx_prefs.gtk.showjoin, gtkhx_prefs.showjoin);
	gtk_table_attach(GTK_TABLE(table), gtkhx_prefs.gtk.showjoin, 0, 1, 3, 4,
			 (GtkAttachOptions)(GTK_FILL),
			 (GtkAttachOptions)(GTK_FILL), 0, 0);

	gtkhx_prefs.gtk.track_case = gtk_check_button_new_with_label(_("Case Sensitive Tracker Searching"));
	gtk_toggle_button_set_active    ((GtkToggleButton *)gtkhx_prefs.gtk.track_case, gtkhx_prefs.track_case);
	gtk_table_attach(GTK_TABLE(table), gtkhx_prefs.gtk.track_case, 0, 1, 4, 5,
			 (GtkAttachOptions)(GTK_FILL),
			 (GtkAttachOptions)(GTK_FILL), 0, 0);

	gtkhx_prefs.gtk.old_nickcompletion = gtk_check_button_new_with_label(_("Use old-style nick completion"));
	gtk_toggle_button_set_active    ((GtkToggleButton *)gtkhx_prefs.gtk.old_nickcompletion, gtkhx_prefs.old_nickcompletion);
	gtk_table_attach(GTK_TABLE(table), gtkhx_prefs.gtk.old_nickcompletion, 0, 1, 5, 6,
			 (GtkAttachOptions)(GTK_FILL),
			 (GtkAttachOptions)(GTK_FILL), 0, 0);

	gtk_box_pack_start(GTK_BOX(wid), table, 0, 0, 0);
}

static void settings_page_general(GtkWidget *vbox)
{
	GtkWidget *wid;
	GtkWidget *table;
	GtkWidget *name;

	wid = settings_create_group(vbox, _("General"));

	table = gtk_table_new(2, 2, 0);
	gtk_table_set_row_spacings(GTK_TABLE(table), 10);
	gtk_table_set_col_spacings(GTK_TABLE(table), 5);

	gtkhx_prefs.gtk.nick = gtk_entry_new();
	gtk_table_attach(GTK_TABLE(table), gtkhx_prefs.gtk.nick, 1, 2, 0, 1,
			 (GtkAttachOptions)(GTK_EXPAND | GTK_FILL),
			 (GtkAttachOptions)0, 0, 0);
	gtk_entry_set_text(GTK_ENTRY(gtkhx_prefs.gtk.nick), sessions[0].htlc.name);

	name = gtk_label_new(_("Your Name:"));
	gtk_table_attach(GTK_TABLE(table), name, 0, 1, 0, 1,
	                  (GtkAttachOptions)(GTK_FILL),
	                  (GtkAttachOptions)(GTK_FILL), 0, 0);
	gtk_label_set_justify(GTK_LABEL(name), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment(GTK_MISC(name), 0, 0.5);

	gtk_box_pack_start(GTK_BOX(wid), table, 0, 0, 0);
}

static void
settings_ctree_select (GtkWidget * ctree, GtkCTreeNode * node)
{
	GtkWidget *book;
	gint page;

	if (!GTK_CLIST (ctree)->selection)
		return;

	book = GTK_WIDGET (gtk_object_get_user_data (GTK_OBJECT (ctree)));
	page = (gint) gtk_ctree_node_get_row_data (GTK_CTREE (ctree), node);

	gtk_notebook_set_page (GTK_NOTEBOOK (book), page);
}

static GtkWidget *
settings_create_page (GtkWidget * book, gchar * book_label, GtkWidget * ctree,
							 gchar * tree_label, GtkCTreeNode * parent,
							 GtkCTreeNode ** node, gint page_index,
							 void (*draw_func) (GtkWidget *))
{
	GtkWidget *frame;
	GtkWidget *label;
	GtkWidget *vvbox;
	GtkWidget *vbox;

	gchar *titles[1];

	vvbox = gtk_vbox_new (FALSE, 0);

	/* border for the label */
	frame = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT);
	gtk_box_pack_start (GTK_BOX (vvbox), frame, FALSE, TRUE, 0);

	/* label */
	label = gtk_label_new (book_label);
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_misc_set_padding (GTK_MISC (label), 2, 1);
	gtk_container_add (GTK_CONTAINER (frame), label);

	/* vbox for the tab */
	vbox = gtk_vbox_new (FALSE, 2);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 4);
	gtk_container_add (GTK_CONTAINER (vvbox), vbox);

	/* label on the tree */
	titles[0] = tree_label;
	*node = gtk_ctree_insert_node (GTK_CTREE (ctree), parent, NULL, titles, 0,
											 NULL, NULL, NULL, NULL, FALSE, FALSE);
	gtk_ctree_node_set_row_data (GTK_CTREE (ctree), *node,
										  (gpointer) page_index);

	/* call the draw func if there is one */
	if (draw_func)
		draw_func (vbox);


	gtk_notebook_append_page (GTK_NOTEBOOK (book), vvbox, NULL);
	return vbox;
}

void create_options_window(GtkWidget *widget, gpointer data)
{
	GtkCTreeNode *last_top;
	GtkCTreeNode *last_child;
	GtkWidget *dialog;
	GtkWidget *hbbox;
	GtkWidget *frame;
	GtkWidget *ctree;
	GtkWidget *book;
	GtkWidget *hbox;
	GtkWidget *vbox;
	GtkWidget *wid;
	gchar *titles[1];
	gint page_index;
	session *sess = data;

	if(gtkhx_prefs.gtk.options_window) {
		gdk_window_raise(gtkhx_prefs.gtk.options_window->window);
		return;
	}

	dialog = gtk_dialog_new ();
	gtk_window_set_wmclass(GTK_WINDOW(dialog), "options", "GtkHx");
	gtk_window_set_title (GTK_WINDOW (dialog), "GtkHx Preferences");
	gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_CENTER);
	gtk_widget_set_usize(dialog, 570, 400);
	gtk_object_set_data(GTK_OBJECT(dialog), "sess", sess);
	gtk_signal_connect_object (GTK_OBJECT (dialog), "delete_event",
							   GTK_SIGNAL_FUNC (close_options_window), 0);

	gtkhx_prefs.gtk.options_window = dialog;

	gtk_container_set_border_width
		(GTK_CONTAINER (GTK_DIALOG (dialog)->action_area), 2);
	gtk_box_set_homogeneous (GTK_BOX (GTK_DIALOG (dialog)->action_area),
									 FALSE);

	hbbox = gtk_hbutton_box_new ();
	gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbbox), 4);
	gtk_box_pack_end (GTK_BOX (GTK_DIALOG (dialog)->action_area), hbbox,
							FALSE, FALSE, 0);

	wid = gtk_button_new_with_label (_("OK"));
	gtk_signal_connect (GTK_OBJECT (wid), "clicked",
						GTK_SIGNAL_FUNC (options_change), GINT_TO_POINTER(0));
	gtk_box_pack_start (GTK_BOX (hbbox), wid, 0, 0, 0);

	wid = gtk_button_new_with_label (_("Apply"));
	gtk_signal_connect (GTK_OBJECT (wid), "clicked",
						GTK_SIGNAL_FUNC (options_change), GINT_TO_POINTER(1));
	gtk_box_pack_start (GTK_BOX (hbbox), wid, 0, 0, 0);

	wid = gtk_button_new_with_label (_("Cancel"));
	gtk_signal_connect (GTK_OBJECT (wid), "clicked",
							  GTK_SIGNAL_FUNC (close_options_window), 0);
	gtk_box_pack_start (GTK_BOX (hbbox), wid, 0, 0, 0);

	hbox = gtk_hbox_new (0, 6);
	gtk_container_set_border_width (GTK_CONTAINER (hbox), 6);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox),
							  hbox, TRUE, TRUE, 0);

	titles[0] = _("Categories");
	ctree = gtk_ctree_new_with_titles (1, 0, titles);
	gtk_clist_set_selection_mode (GTK_CLIST (ctree), GTK_SELECTION_BROWSE);
	gtk_ctree_set_indent (GTK_CTREE (ctree), 15);
	gtk_widget_set_usize (ctree, 140, 0);
	gtk_box_pack_start (GTK_BOX (hbox), ctree, 0, 0, 0);

	frame = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
	gtk_box_pack_start (GTK_BOX (hbox), frame, TRUE, TRUE, 0);

	book = gtk_notebook_new ();
	gtk_notebook_set_show_tabs (GTK_NOTEBOOK (book), FALSE);
	gtk_notebook_set_show_border (GTK_NOTEBOOK (book), FALSE);
	gtk_container_add (GTK_CONTAINER (frame), book);
	gtk_object_set_user_data (GTK_OBJECT (ctree), book);
	gtk_signal_connect (GTK_OBJECT (ctree), "tree_select_row",
							  GTK_SIGNAL_FUNC (settings_ctree_select), NULL);
	page_index = 0;

	vbox = settings_create_page(book, _("General Settings"), ctree, _("General Settings"), 0, &last_top, page_index++, settings_page_general);
	gtk_ctree_select (GTK_CTREE (ctree), last_top);

	vbox = settings_create_page(book, _("Tracker Settings"), ctree, _("Trackers"), last_top, &last_child, page_index++, settings_page_tracker);

	vbox = settings_create_page(book, _("Icon Settings"), ctree, _("Icons"), last_top, &last_child, page_index++, settings_page_icon);

	vbox = settings_create_page(book, _("Path Settings"), ctree, _("Paths"), last_top, &last_child, page_index++, settings_page_path);

	vbox = settings_create_page(book, _("Files Settings"), ctree, _("Files"), last_top, &last_child, page_index++, settings_page_files);

	vbox = settings_create_page(book, _("Miscellaeneous"), ctree, _("Miscellaeneous"), last_top, &last_child, page_index++, settings_page_misc);

	vbox = settings_create_page(book, _("XText Settings"), ctree, _("XText Settings"), 0, &last_top, page_index++, settings_page_xtext);

	vbox = settings_create_page(book, _("Font Settings"), ctree, _("Fonts"), last_top, &last_child, page_index++, settings_page_font);

	vbox = settings_create_page(book, _("Sound Settings"), ctree, _("Sound Settings"), 0, &last_top, page_index++, settings_page_sound);

	gtk_ctree_expand_recursive ((GtkCTree *) ctree, 0);
	gtk_clist_select_row (GTK_CLIST (ctree), 0, 0);

	gtk_widget_show_all(dialog);

	list_icons();
}
