/*
 * Copyright (C) 2001 Misha Nasledov <misha@nasledov.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <gtk/gtk.h>
#include <sys/time.h>
#include <time.h>
#include <netinet/in.h>
#include "hotline.h"
#include "hxd.h"
#include "hx.h"
#include "network.h"
#include "xfers.h"
#include "chat.h"
#include "tasks.h"
#include "files.h"
#include "gtkutil.h"
#include "msg.h"
#include "news.h"
#include "sound.h"
#include "plugin.h"
#include "users.h"
#include "usermod.h"
#include "news15.h"

size_t news_len = 0;
guint8 *news_buf = 0;
char *hx_timeformat = "%c";
extern int xfer_go_timer (void *__arg);

void rcv_task_user_list (struct htlc_conn *htlc, struct chat *chat, int text);

void hx_rcv_chat (struct htlc_conn *htlc)
{
	guint32 cid = 0;
	guint16 len = 0;
	char chatbuf[8192 + 1], *chat;
	guint32 uid = 0;
	session *sess = sess_from_htlc(htlc);
	struct chat *hx_chat = chat_with_cid(sess, 0);

	dh_start(htlc)
		switch (ntohs(dh->type)) {
		case HTLS_DATA_CHAT:
			len = (ntohs(dh->len) > sizeof(chatbuf) - 1 ? sizeof(chatbuf) - 1 : ntohs(dh->len));
			memcpy(chatbuf, dh->data, len);
			break;
		case HTLS_DATA_CHAT_ID:
			dh_getint(cid);
			break;
		case HTLS_DATA_UID: /* this only works on hxd 0.1.35 or higher */
			dh_getint(uid);
			break;
		}
	dh_end()

	if(uid) { /* do ignoring stuff */
		struct hx_user *user = hx_user_with_uid(hx_chat->user_list, uid);
		if(user && user->ignore) {
			return;
		}
	}

	CR2LF(chatbuf, len);
	strip_ansi(chatbuf, len);
	chatbuf[len] = 0;

	if (chatbuf[0] == '\n') {
		chat = chatbuf+1;
		len--;
	} else {
		chat = chatbuf;
	}

#ifdef USE_PLUGIN
	if(EMIT_SIGNAL(XP_RCV_CHAT, sess, chat, &cid, &uid, 0, 0) == 1) {
		return;
	}
#endif

	hx_output.chat(htlc, cid, chat, len);
	play_sound(CHAT_POST);
}

void hx_rcv_msg (struct htlc_conn *htlc)
{
	guint32 uid = 0;
	guint16 msglen = 0, nlen = 0;
	char msgbuf[8192 + 1], name[128 + 1];
	session *sess = sess_from_htlc(htlc);
	struct chat *chat = chat_with_cid(sess, 0);
	struct hx_user *user = 0;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_UID:
				dh_getint(uid);
				break;
			case HTLS_DATA_MSG:
				msglen = ntohs(dh->len) > 8192 ? 8192 : ntohs(dh->len);
				memcpy(msgbuf, dh->data, msglen);
				break;
			case HTLS_DATA_NAME:
				nlen = ntohs(dh->len) > 128 ? 128 : ntohs(dh->len);
				memcpy(name, dh->data, nlen);
				strip_ansi(name, nlen);
				break;
		}
	dh_end()


	user = hx_user_with_uid(chat->user_list, uid);
	if(user && user->ignore) {
		return;
	}

	CR2LF(msgbuf, msglen);
	strip_ansi(msgbuf, msglen);
	name[nlen] = 0;
	msgbuf[msglen] = 0;

#ifdef USE_PLUGIN
	if(EMIT_SIGNAL(XP_RCV_MSG, sess, msgbuf, name, &uid, 0, 0) == 1) {
		return;
	}
#endif

	if(uid > 0) {
		hx_output.msg(name, uid, msgbuf);
	}
	else {
		broadcastmsg(msgbuf);
	}
	play_sound(MSG);

	if (!*last_msg_nick) {
		strncpy(last_msg_nick, name, 31);
		last_msg_nick[31] = 0;
	}
}

void hx_rcv_agreement_file (struct htlc_conn *htlc)
{
	gint16 len = 0;

	dh_start(htlc)
		if (ntohs(dh->type) != HTLS_DATA_AGREEMENT)
			continue;
	len = ntohs(dh->len);
	CR2LF(dh->data, len);
	strip_ansi(dh->data, len);
#ifdef USE_PLUGIN
	if(EMIT_SIGNAL(XP_RCV_AGREE, sess_from_htlc(htlc), dh->data, &len, 0, 0, 0) == 1) {
		return;
	}
#endif
	hx_output.agreement(sess_from_htlc(htlc), dh->data, len);

	dh_end()
}

void hx_rcv_news_post (struct htlc_conn *htlc)
{
	dh_start(htlc)
		if (ntohs(dh->type) != HTLS_DATA_NEWS)
			continue;
		news_len += ntohs(dh->len);
		news_buf = g_realloc(news_buf, news_len + 1);
		memmove(&(news_buf[ntohs(dh->len)]), news_buf, news_len - ntohs(dh->len));
		memcpy(news_buf, dh->data, ntohs(dh->len));
		CR2LF(news_buf, ntohs(dh->len));
		strip_ansi(news_buf, ntohs(dh->len));
		news_buf[news_len] = 0;
		hx_output.news_post(htlc, news_buf, ntohs(dh->len));
		play_sound(NEWS_POST);
	dh_end()
}

void hx_rcv_task (struct htlc_conn *htlc)
{
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
	guint32 trans = ntohl(h->trans);
	struct task *tsk = task_with_trans(sess_from_htlc(htlc), trans);

	if ((ntohs(h->flag) & 1))
		task_error(htlc);
	if (tsk) {
		/* XXX tsk->rcv might call task_delete */
		int fd = htlc->fd;
		if (tsk->rcv)
			tsk->rcv(htlc, tsk->ptr, tsk->data);
		if (hxd_files[fd].conn.htlc)
			task_delete(sess_from_htlc(htlc), tsk);
	} else {
		/*	hx_printf_prefix(0, INFOPREFIX, "got task 0x%08x\n", trans); */
	}
}

void hx_rcv_user_change (struct htlc_conn *htlc)
{
	guint32 uid = 0, cid = 0, icon = 0, color = 0;
	guint16 nlen = 0, got_color = 0;
	guint8 name[32];
	struct chat *chat;
	struct hx_user *user;
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
	session *sess = sess_from_htlc(htlc);

	if ((ntohs(h->flag) & 1))
		return;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_UID:
				dh_getint(uid);
				break;
			case HTLS_DATA_ICON:
				dh_getint(icon);
				break;
			case HTLS_DATA_NAME:
				nlen = ntohs(dh->len) > 31 ? 31 : ntohs(dh->len);
				memcpy(name, dh->data, nlen);
				name[nlen] = 0;
				strip_ansi(name, nlen);
				break;
			case HTLS_DATA_COLOUR:
				dh_getint(color);
				got_color = 1;
				break;
			case HTLS_DATA_CHAT_ID:
				dh_getint(cid);
				break;
		}
	dh_end()
		chat = chat_with_cid(sess, cid);
	if (!chat) {
		chat = chat_new(sess, cid);
	}
	user = hx_user_with_uid(chat->user_list, uid);
	if (!user) {
		user = hx_user_new(&chat->user_tail);
		chat->nusers++;
		user->uid = uid;
		hx_output.user_create(htlc, chat, user, name, icon, color);
		play_sound(USER_JOIN);
		if(gtkhx_prefs.showjoin) {
			hx_printf_prefix(htlc, cid, INFOPREFIX, "join: %s\n", name);
		}
	}

	else {


		if (!got_color)
			color = user->color;
		hx_output.user_change(htlc, chat, user, name, icon, color);
		if((user->color == color && user->icon == icon) || strcmp(name, user->name)) {
			if(user->ignore) return;
			hx_printf_prefix(htlc, cid, INFOPREFIX, "%s is now known as %s\n", user->name, name);
		}

	}
	if (nlen) {
		memcpy(user->name, name, nlen);
		user->name[nlen] = 0;
	}
	if (icon)
		user->icon = icon;
	if (got_color)
		user->color = color;
	if (uid && uid == htlc->uid) {
		htlc->uid = user->uid;
		htlc->icon = user->icon;
		htlc->color = user->color;
		strcpy(htlc->name, user->name);
	}
}


void hx_rcv_user_part (struct htlc_conn *htlc)
{
	guint32 uid = 0, cid = 0;
	struct chat *chat;
	struct hx_user *user;
	char *col;
	session *sess = sess_from_htlc(htlc);

	dh_start(htlc)
		switch (ntohs(dh->type)) {
		case HTLS_DATA_UID:
			dh_getint(uid);
			break;
		case HTLS_DATA_CHAT_ID:
			dh_getint(cid);
			break;
		}
	dh_end()
		chat = chat_with_cid(sess, cid);
	if (!chat)
		return;
	user = hx_user_with_uid(chat->user_list, uid);
	if (user) {
		col = colorstr(user->color);
		hx_output.user_delete(htlc, chat, user);
		hx_user_delete(&chat->user_tail, user);
		chat->nusers--;
		play_sound(USER_PART);
		if(gtkhx_prefs.showjoin) {
			hx_printf_prefix(htlc, cid, INFOPREFIX, "parts: %s \n", user->name);
		}
	}
}

void hx_rcv_chat_subject (struct htlc_conn *htlc)
{
	guint32 cid = 0;
	guint16 slen = 0;
	guint8 subject[256];
	struct chat *chat;
	session *sess = sess_from_htlc(htlc);

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_CHAT_ID:
				dh_getint(cid);
				break;
			case HTLS_DATA_CHAT_SUBJECT:
				slen = ntohs(dh->len) > 255 ? 255 : ntohs(dh->len);
				memcpy(subject, dh->data, slen);
				break;
		}
	dh_end()
	if (slen) {
		chat = chat_with_cid(sess, cid);
		if (!chat)
			return;
		if(strcmp(subject, chat->subject) == 0)
			return;
		memcpy(chat->subject, subject, slen);
		chat->subject[slen] = 0;

#ifdef USE_PLUGIN
		if(EMIT_SIGNAL(XP_RCV_SUBJ, sess, chat->subject, &cid, 0, 0, 0) == 1) {
			return;
		}
#endif
		hx_output.chat_subject(htlc, cid, chat->subject);
	}
}

void hx_rcv_chat_invite (struct htlc_conn *htlc)
{
	guint32 uid = 0, cid = 0;
	guint16 nlen;
	guint8 name[32];
	session *sess = sess_from_htlc(htlc);
	struct chat *chat = chat_with_cid(sess, 0);
	struct hx_user *user = 0;

	name[0] = 0;
	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_UID:
				dh_getint(uid);
				break;
			case HTLS_DATA_CHAT_ID:
				dh_getint(cid);
				break;
			case HTLS_DATA_NAME:
				nlen = ntohs(dh->len) > 31 ? 31 : ntohs(dh->len);
				memcpy(name, dh->data, nlen);
				name[nlen] = 0;
				strip_ansi(name, nlen);
				break;
		}
	dh_end()

	user = hx_user_with_uid(chat->user_list, uid);
	if(user && user->ignore) {
		return;
	}
#ifdef USE_PLUGIN
	if(EMIT_SIGNAL(XP_RCV_INVITE, sess, name, &uid, &cid, 0, 0) == 1) {
		return;
	}
#endif
	hx_output.chat_invitation(htlc, cid, name);
	play_sound(CHAT_INVITE);
}

void hx_rcv_user_selfinfo (struct htlc_conn *htlc)
{
	struct hl_userlist_hdr *uh;
	guint16 nlen;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_ACCESS:
				if (ntohs(dh->len) != 8)
					break;
				memcpy(&htlc->access, dh->data, 8);
				break;
			case HTLS_DATA_USER_LIST:
				if (ntohs(dh->len) < (SIZEOF_HL_USERLIST_HDR - SIZEOF_HL_DATA_HDR))
					break;
				uh = (struct hl_userlist_hdr *)dh;
				htlc->uid = ntohs(uh->uid);
				htlc->icon = ntohs(uh->icon);
				htlc->color = ntohs(uh->color);
				nlen = ntohs(uh->nlen) > 31 ? 31 : ntohs(uh->nlen);
				memcpy(htlc->name, uh->name, nlen);
				htlc->name[nlen] = 0;
				break;
		}
	dh_end()

		setbtns(&sessions[0], 1);
}

void hx_rcv_dump (struct htlc_conn *htlc)
{
	int fd;

	fd = open("hx.dump", O_WRONLY|O_APPEND|O_CREAT, 0644);
	if (fd < 0)
		return;
	write(fd, htlc->in.buf, htlc->in.pos);
	fsync(fd);
	close(fd);
}

void hx_rcv_xfer_queue(struct htlc_conn *htlc)
{
	guint16 queueid = 0;
	guint16 ref = 0;
	struct htxf_conn *htxf;

	dh_start(htlc) {
		switch(dh->type) {
		case HTLS_DATA_HTXF_REF:
			dh_getint(ref);
			break;
		case HTLS_DATA_QUEUE:
			dh_getint(queueid);
			break;
		}
	} dh_end()

	htxf = htxf_with_ref(ref);
	if(!htxf) {
		return;
	}
	htxf->queue = queueid;
	hx_output.xfer_queue(htxf);
}



void hx_rcv_hdr (struct htlc_conn *htlc)
{
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
	guint32 len;

	if (ntohl(h->len) < 2) {
		len = 0;
	}
	else {
		len = (ntohl(h->len) > MAX_HOTLINE_PACKET_LEN ? MAX_HOTLINE_PACKET_LEN : ntohl(h->len)) - 2;
	}

	/* htlc->trans = ntohl(h->trans); */
	htlc->rcv = 0;
	switch (ntohl(h->type)) {
	case HTLS_HDR_CHAT:
		htlc->rcv = hx_rcv_chat;
		break;
	case HTLS_HDR_MSG:
		htlc->rcv = hx_rcv_msg;
		break;
	case HTLS_HDR_USER_CHANGE:
	case HTLS_HDR_CHAT_USER_CHANGE:
		htlc->rcv = hx_rcv_user_change;
		break;
	case HTLS_HDR_USER_PART:
	case HTLS_HDR_CHAT_USER_PART:
		htlc->rcv = hx_rcv_user_part;
		break;
	case HTLS_HDR_NEWS_POST:
		htlc->rcv = hx_rcv_news_post;
		break;
	case HTLS_HDR_TASK:
		htlc->rcv = hx_rcv_task;
		break;
	case HTLS_HDR_CHAT_SUBJECT:
		htlc->rcv = hx_rcv_chat_subject;
		break;
	case HTLS_HDR_CHAT_INVITE:
		htlc->rcv = hx_rcv_chat_invite;
		break;
	case HTLS_HDR_MSG_BROADCAST:
		htlc->rcv = hx_rcv_msg;
		break;
	case HTLS_HDR_USER_SELFINFO:
		htlc->rcv = hx_rcv_user_selfinfo;
		break;
	case HTLS_HDR_AGREEMENT:
		htlc->rcv = hx_rcv_agreement_file;
		break;
	case HTLS_HDR_POLITEQUIT:
		hx_printf_prefix(htlc, 0, INFOPREFIX, "polite quit\n");
		htlc->rcv = hx_rcv_msg;
		break;
	case HTLS_HDR_QUEUE:
		htlc->rcv = hx_rcv_xfer_queue;
		break;
	default:
		g_print("0x%08x\n", ntohl(h->type));
		hx_printf_prefix(htlc, 0, INFOPREFIX, "unknown header type 0x%08x\n", ntohl(h->type));
		htlc->rcv = hx_rcv_dump;
		break;
	}

	if (len) {
		qbuf_set(&htlc->in, htlc->in.pos, len);
	} else {
		if (htlc->rcv)
			htlc->rcv(htlc);
		htlc->rcv = hx_rcv_hdr;
		qbuf_set(&htlc->in, 0, SIZEOF_HL_HDR);
	}
}

void rcv_task_user_open (struct htlc_conn *htlc, struct uesp_fn *uespfn)
{
	char name[32], login[32], pass[32];
	guint16 nlen = 0, llen = 0, plen = 0;
	struct hl_access_bits *access = 0;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
		case HTLS_DATA_NAME:
			if(ntohs(dh->len) >= sizeof(name))
				nlen = sizeof(name)-1;
			else
				nlen = ntohs(dh->len);
			memcpy(name, dh->data, nlen);
			break;
		case HTLS_DATA_LOGIN:
			if(ntohs(dh->len) >= sizeof(login))
				llen = sizeof(login)-1;
			else
				llen = ntohs(dh->len);
			hl_decode(login, dh->data, llen);
			break;
		case HTLS_DATA_PASSWORD:
			if(ntohs(dh->len) >= sizeof(pass))
				plen = sizeof(pass)-1;
			else
				plen = ntohs(dh->len);
			if (plen > 1 && dh->data[0])
				hl_decode(pass, dh->data, plen);
			else
				pass[0] = 0;
			break;
		case HTLS_DATA_ACCESS:
			if (ntohs(dh->len) >= 8)
				access = (struct hl_access_bits *)dh->data;
			break;
		}
	dh_end()
	name[nlen] = 0;
	login[llen] = 0;
	pass[plen] = 0;
	if (access)
		uespfn->fn(uespfn->uesp, name, login, pass, access);
	g_free(uespfn);
}


void rcv_task_msg (struct htlc_conn *htlc __attribute__((__unused__)), char *msg_buf)
{
	if (msg_buf) {
		hx_printf(htlc, 0, "%s\n", msg_buf);
		g_free(msg_buf);
	}
}

void rcv_task_newscat_list(struct htlc_conn *htlc, char *path)
{
	guint32 len;
	struct news_group *group = g_malloc(sizeof(struct news_group));
	struct news_item *ni;
	unsigned char *ptr;
	guint16 u16;
	guint32 u32;
	int p, j;
	int exit = 0;

	/* Taken from fidelio =) */
#define get_pstring(ret) if(*ptr==0){ret=NULL;}else{ret=g_malloc(1+*ptr);memcpy(ret,ptr+1,*ptr);(ret)[*ptr]=0;ptr+=*ptr;} ptr++
#define get_short(ret) memcpy(&u16,ptr,2);ptr+=2;ret=GUINT16_FROM_BE(u16)
#define get_long(ret) memcpy(&u32,ptr,4);ptr+=4;ret=GUINT32_FROM_BE(u32)

	group->path = g_malloc(strlen(path));
	strcpy(group->path, path);
	g_free(path);

	dh_start(htlc)
		switch (ntohs(dh->type)) {
		case HTLC_DATA_CATLIST:
			len = ntohs(dh->len);
			ptr = dh->data+4;
			get_long(group->post_count);
			ptr += *ptr;
			ptr += 2;
			if(!group->post_count) {
				exit = 1;
				break;
			}
			group->posts = g_malloc(sizeof(struct news_item)*group->post_count);
			for(p = 0, ni = group->posts; p < group->post_count; p++, ni++) {
				get_long(ni->postid);
				get_short(ni->date.base_year);
				get_short(ni->date.pad);
				get_long(ni->date.seconds);
				get_long(ni->parentid);
				ptr += 4;
				get_short(ni->partcount);
				get_pstring(ni->subject);
				get_pstring(ni->sender);
				ni->parts = g_malloc(sizeof(ni->parts[0])*ni->partcount);
				ni->size = 0;
				for(j=0;j<ni->partcount;j++){
					get_pstring(ni->parts[j].mime_type);
					get_short(ni->parts[j].size);
					ni->size+=ni->parts[j].size;
				}
				ni->group = group;
			}
			break;
		}

	dh_end()

	if(exit) {
		return;
	}

	hx_output.news_catalog(group);
}

void rcv_task_newsfolder_list(struct htlc_conn *htlc, char *path)
{
	guint32 len;
	struct news_folder *folder = g_malloc(sizeof(struct news_folder));
	struct folder_item *item;
	int num = 0;

	folder->entry = g_malloc(sizeof(struct folder_item));
	folder->path = path;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
		case HTLC_DATA_NEWSFOLDERITEM:
			len = ntohs(dh->len);
			num++;
			folder->entry = g_realloc(folder->entry, sizeof(struct folder_item)*num);
			item = g_malloc(sizeof(struct folder_item));
			item->type = dh->data[0];
			item->name = g_malloc(len);
			memcpy(item->name, dh->data+1, len-1);
			item->name[len-1] = 0;
			folder->entry[num-1] = item;
			break;
		}
	dh_end()

	folder->num_entries = num;
	hx_output.news_folder(folder);
}

void rcv_task_news_post (struct htlc_conn *htlc, struct news_item *item)
{
	guint32 len;
	struct news_post *post = 0;
	guint32 postid;

	dh_start(htlc)
		switch(ntohs(dh->type)) {
		case HTLC_DATA_NEWSDATA:
			len = ntohs(dh->len);
			post = g_malloc(sizeof(struct news_post));
			post->buf = g_malloc(len);
			memcpy(post->buf, dh->data, len);
			CR2LF(post->buf, len);
			strip_ansi(post->buf, len);
			post->buf[len] = 0;
			break;
		case HTLC_DATA_THREADID:
			dh_getint(postid);
			break;
		case HTLS_DATA_TASKERROR:
			return;
		}
	dh_end()

	post->item = item;
	hx_output.news_thread(post);
}


void rcv_task_news_users(struct htlc_conn *htlc, struct chat *chat, int text)
{

	/* output user list and then grab news */
	/* this is only used for login events  */
	rcv_task_user_list(htlc, chat, text);

	reload_news(0, sess_from_htlc(htlc));

}

void rcv_task_login (struct htlc_conn *htlc, char *pass __attribute__((__unused__)))
{
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
	char buf[HOSTLEN];
	guint32 uid;
	guint16 version;
	guint16 len;
	char servername[8192+1];
	session *sess = sess_from_htlc(htlc);


#ifdef USE_IPV6
	getnameinfo(htlc->addr->ai_addr, htlc->addr->ai_addrlen, buf, sizeof(buf), NULL, 0, NI_NUMERICHOST);

#else
	inet_ntop(AF_INET, &htlc->addr.sin_addr, buf, sizeof(buf));
#endif
	hx_printf_prefix(htlc, 0, INFOPREFIX, "%s:%u: %s %s\n",
					 buf,

#ifdef USE_IPV6
					 server_port,

#else
					 ntohs(htlc->addr.sin_port),
#endif
					 _("login"),

					 (ntohs(h->flag) & 1) ? _("failed?") : _("successful"));


	if (!(ntohs(h->flag) & 1)) {
		play_sound(LOGIN);
		changetitlesconnected(sess);
		setbtns(sess, 1);
		set_status_bar(2);
		connected = 1;

		dh_start(htlc)
			switch (ntohs(dh->type)) {
			case HTLS_DATA_UID:
				dh_getint(uid);
				htlc->uid = uid;
				break;
			case HTLS_DATA_VERSION: /* Hotline 1.5+ servers only */
				dh_getint(version);
				htlc->version = version;
				break;
			case HTLS_DATA_SERVERNAME: /* Hotline 1.5+ servers only */
				len = (ntohs(dh->len) > sizeof(servername) - 1 ? sizeof(servername) - 1 : ntohs(dh->len));
				memcpy(servername, dh->data, len);
				CR2LF(servername, len);
				strip_ansi(servername, len);
				servername[len] = 0;
				server_addr = g_strdup(servername);
				changetitlesconnected(sess);
				break;
			}
		dh_end()


		/* this will get news and users */
		task_new(htlc, rcv_task_news_users, sess->chat_list, 0, "who");
		hlwrite(htlc, HTLC_HDR_USER_GETLIST, 0, 0);
	}
}

void rcv_task_news_file (struct htlc_conn *htlc)
{
	dh_start(htlc) {
		if (ntohs(dh->type) != HTLS_DATA_NEWS)
			continue;
		news_len = ntohs(dh->len);
		news_buf = g_realloc(news_buf, news_len + 1);
		memcpy(news_buf, dh->data, news_len);
		CR2LF(news_buf, news_len);
		strip_ansi(news_buf, news_len);
		news_buf[news_len] = 0;
	dh_end()

	}

	hx_output.news_file(htlc, news_buf, news_len);
}

void rcv_task_user_list (struct htlc_conn *htlc, struct chat *chat, int text)
{
	struct hl_userlist_hdr *uh;
	struct hx_user *user;
	guint16 nlen;
	int new = 0;
	struct gtkhx_chat *gchat = gchat_with_cid(sess_from_htlc(htlc), chat->cid);

	dh_start(htlc)
		if (ntohs(dh->type) == HTLS_DATA_USER_LIST) {
			uh = (struct hl_userlist_hdr *)dh;
			user = hx_user_with_uid(chat->user_list, ntohs(uh->uid));
			if (!user) {
				new = 1;
				user = hx_user_new(&chat->user_tail);
				chat->nusers++;
			}
			user->uid = ntohs(uh->uid);
			user->icon = ntohs(uh->icon);
			user->color = ntohs(uh->color);
			nlen = ntohs(uh->nlen) > 31 ? 31 : ntohs(uh->nlen);
			memcpy(user->name, uh->name, nlen);
			strip_ansi(user->name, nlen);
			user->name[nlen] = 0;
			if (!htlc->uid && !strcmp(user->name, htlc->name) &&

				user->icon == htlc->icon) {
				htlc->uid = user->uid;
				htlc->color = user->color;
			}
			if (new) {
				hx_output.user_create(htlc, chat, user, user->name, user->icon,
									  user->color);
			}
		}

		else if (ntohs(dh->type) == HTLS_DATA_CHAT_SUBJECT) {
			guint16 slen = ntohs(dh->len) > 255 ? 255 : ntohs(dh->len);
			memcpy(chat->subject, dh->data, slen);
			chat->subject[slen] = 0;
			if(gchat && gchat->subject) {
				gtk_entry_set_text(GTK_ENTRY(gchat->subject), chat->subject);
			}
		}
	dh_end()
}

void rcv_task_user_list_switch (struct htlc_conn *htlc, struct chat *chat)
{
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
	session *sess = sess_from_htlc(htlc);

	if ((ntohs(h->flag) & 1)) {
		chat_delete(sess, chat);
		return;
	}

	rcv_task_user_list(htlc, chat, 0);
}

void rcv_task_kick (struct htlc_conn *htlc)
{
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;

	if ((ntohs(h->flag) & 1))
		return;
	hx_printf_prefix(htlc, 0, INFOPREFIX, "%s\n", _("kick successful"));
}


void rcv_task_user_info (struct htlc_conn *htlc, guint32 uid, int text)
{
	guint16 ilen = 0, nlen = 0;
	guint8 info[4096 + 1], name[32];

	name[0] = 0;
	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_USER_INFO:
				ilen = ntohs(dh->len) > 4096 ? 4096 : ntohs(dh->len);
				memcpy(info, dh->data, ilen);
				info[ilen] = 0;
				break;
			case HTLS_DATA_NAME:
				nlen = ntohs(dh->len) > 31 ? 31 : ntohs(dh->len);
				memcpy(name, dh->data, nlen);
				name[nlen] = 0;
				strip_ansi(name, nlen);
				break;
		}
	dh_end()
	if (nlen && ilen) {
		CR2LF(info, ilen);
		strip_ansi(info, ilen);
		hx_output.user_info(uid, name, info, ilen);
	}
}

void rcv_task_file_list (struct htlc_conn *htlc, struct cached_filelist *cfl, void *data)
{
	struct hl_filelist_hdr *fh = 0;
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
	guint32 fh_len;

	if ((ntohs(h->flag) & 1)) {
		cfl_delete(cfl);
		return;
	}
	dh_start(htlc)
		if (ntohs(dh->type) != HTLS_DATA_FILE_LIST)
			continue;
		fh = (struct hl_filelist_hdr *)dh;
		if (cfl->completing > 1) {
			char *pathbuf;
			int fnlen, len;

			fnlen = ntohl(fh->fnlen);
			len = strlen(cfl->path) + 1 + fnlen + 1;
			pathbuf = g_malloc(len);
			snprintf(pathbuf, len, "%s%c%.*s", cfl->path[1] ? cfl->path : "", dir_char, (int)fnlen, fh->fname);
			if (ntohl(fh->ftype) == 0x666c6472) {
				struct cached_filelist *ncfl;
				guint16 hldirlen;
				guint8 *hldir;

				ncfl = cfl_lookup(pathbuf);
				ncfl->completing = cfl->completing;
				ncfl->filter_argv = cfl->filter_argv;
				if (!ncfl->path)
					ncfl->path = g_strdup(pathbuf);
				hldir = path_to_hldir(pathbuf, &hldirlen, 0);
				task_new(&sessions[0].htlc, rcv_task_file_list, ncfl, 0, "ls_complete");
				hlwrite(&sessions[0].htlc, HTLC_HDR_FILE_LIST, 0, 1,
					HTLC_DATA_DIR, hldirlen, hldir);
				g_free(hldir);
			} else if (cfl->completing == COMPLETE_GET_R) {
				struct htxf_conn *htxf;
				char *lpath, *p;

				lpath = g_malloc(len);
				dirmask(lpath, pathbuf, htlc->rootdir);
				p = lpath+1;
				while ((p = strchr(p, dir_char))) {
					*p = 0;
					if (mkdir(lpath+1, S_IRUSR|S_IWUSR|S_IXUSR)) {
						if (errno != EEXIST)
							hx_printf_prefix(htlc, 0, INFOPREFIX, "mkdir(%s): %s\n", lpath+1, strerror(errno));
					}
					*p++ = '/';
					while (*p == dir_char)
						*p++ = '/';
				}
				p = basename(lpath+1);
				if (p)
					dirchar_fix(p);
				htxf = xfer_new(lpath+1, pathbuf, XFER_GET);
				htxf->filter_argv = cfl->filter_argv;
				g_free(lpath);
			}
			g_free(pathbuf);
		}
		fh_len = SIZEOF_HL_DATA_HDR + ntohs(fh->len);
		fh_len += 4 - (fh_len % 4);
		cfl->fh = g_realloc(cfl->fh, cfl->fhlen + fh_len);
		memcpy((char *)cfl->fh + cfl->fhlen, fh, SIZEOF_HL_DATA_HDR + ntohs(fh->len));
		*((guint16 *)((char *)cfl->fh + cfl->fhlen + 2)) = htons(fh_len - SIZEOF_HL_DATA_HDR);
		cfl->fhlen += fh_len;
	dh_end()
	cfl_print(cfl, data);
	cfl->completing = COMPLETE_NONE;
}

void rcv_task_file_getinfo (struct htlc_conn *htlc)
{
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
	guint8 icon[4], type[32], crea[32], date_create[8], date_modify[8];
	guint8 name[256], comment[256];
	guint16 nlen, clen, tlen;
	guint32 size = 0;
	char created[32], modified[32];
	time_t t;
	struct tm tm;

	if ((ntohs(h->flag) & 1))
		return;
	name[0] = comment[0] = type[0] = crea[0] = 0;
	dh_start(htlc)
		switch(ntohs(dh->type)) {
			case HTLS_DATA_FILE_ICON:
				if (ntohs(dh->len) >= 4)
					memcpy(icon, dh->data, 4);
				break;
			case HTLS_DATA_FILE_TYPE:
				tlen = ntohs(dh->len) > 31 ? 31 : ntohs(dh->len);
				memcpy(type, dh->data, tlen);
				type[tlen] = 0;
				break;
			case HTLS_DATA_FILE_CREATOR:
				clen = ntohs(dh->len) > 31 ? 31 : ntohs(dh->len);
				memcpy(crea, dh->data, clen);
				crea[clen] = 0;
				break;
			case HTLS_DATA_FILE_SIZE:
				if (ntohs(dh->len) >= 4)
					size = ntohl(*((guint32 *)dh->data));
				break;
			case HTLS_DATA_FILE_NAME:
				nlen = ntohs(dh->len) > 255 ? 255 : ntohs(dh->len);
				memcpy(name, dh->data, nlen);
				name[nlen] = 0;
				strip_ansi(name, nlen);
				break;
			case HTLS_DATA_FILE_DATE_CREATE:
				if (ntohs(dh->len) >= 8)
					memcpy(date_create, dh->data, 8);
				break;
			case HTLS_DATA_FILE_DATE_MODIFY:
				if (ntohs(dh->len) >= 8)
					memcpy(date_modify, dh->data, 8);
				break;
			case HTLS_DATA_FILE_COMMENT:
				clen = ntohs(dh->len) > 255 ? 255 : ntohs(dh->len);
				memcpy(comment, dh->data, clen);
				comment[clen] = 0;
				CR2LF(comment, clen);
				strip_ansi(name, clen);
				break;
		}
	dh_end()

	t = ntohl(*((guint32 *)&date_create[4])) - 0x7c25b080;
	localtime_r(&t, &tm);
	strftime(created, 31, hx_timeformat, &tm);
	t = ntohl(*((guint32 *)&date_modify[4])) - 0x7c25b080;
	localtime_r(&t, &tm);
	strftime(modified, 31, hx_timeformat, &tm);

	hx_output.file_info(name, crea, type, comment, modified, created, size);
}

void rcv_task_file_get (struct htlc_conn *htlc, struct htxf_conn *htxf)
{
	guint32 ref = 0, size = 0, queue = 0;
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
	int i;
#ifdef USE_IPV6
	int error;
	char portstr[HOSTLEN];
	struct addrinfo hints;

	memset(&hints, 0, sizeof(struct addrinfo));
	hints.ai_family = PF_UNSPEC;
	hints.ai_flags = AI_CANONNAME;
	hints.ai_socktype = SOCK_STREAM;
#endif

	for(i = 0; i < nxfers; i++) {
		if(xfers[i] == htxf) {
			break;
		}
	}

	if(i == nxfers) {
		return;
	}
	if ((ntohs(h->flag) & 1)) {
		if (htxf->opt.retry) {
			htxf->gone = 0;
			timer_add_secs(1, xfer_go_timer, htxf);
		} else {
			gtask_delete_htxf(sess_from_htlc(htlc), htxf);
			xfer_delete(htxf);
		}
		return;
	}

	dh_start(htlc) {
 		switch (ntohs(dh->type)) {
		case HTLS_DATA_HTXF_SIZE:
			dh_getint(size);
			break;
		case HTLS_DATA_HTXF_REF:
			dh_getint(ref);
			break;
		case HTLS_DATA_QUEUE:
			dh_getint(queue); /* Only on 1.5+ servers */
			break;
		}
	}	dh_end()
	if (!size || !ref)
		return;

	htxf->ref = ref;
	htxf->total_size = size;
	htxf->queue = queue;

	gettimeofday(&htxf->start, 0);
#ifdef USE_IPV6
	sprintf(portstr, "%u", server_port+1);

	if((error = getaddrinfo(htlc->addr->ai_canonname, portstr, &hints, &htxf->listen_addr)))
	{
		hx_printf_prefix(&sessions[0].htlc, 0, INFOPREFIX, "htxf %s:%u failed: %s\n", htlc->addr->ai_canonname, server_port+1, gai_strerror(error));
		gtask_delete_htxf(sess_from_htlc(htlc), htxf);
		xfer_delete(htxf);
		return;
	}
#else
	htxf->listen_addr = htlc->addr;
	htxf->listen_addr.sin_port = htons(ntohs(htxf->listen_addr.sin_port)+1);
#endif

	hx_output.xfer_queue(htxf); /* we most certainly want to

								   output its position in the queue */

	xfer_ready_write(htxf);
}

void rcv_task_file_put (struct htlc_conn *htlc, struct htxf_conn *htxf)
{
	guint32 ref = 0, data_pos = 0, rsrc_pos = 0, queue = 0;
	struct stat sb;
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
#ifdef USE_IPV6
	char portstr[HOSTLEN];
	struct addrinfo hints;
	int error;

	memset(&hints, 0, sizeof(struct addrinfo));
	hints.ai_family = PF_UNSPEC;
	hints.ai_flags = AI_CANONNAME;
	hints.ai_socktype = SOCK_STREAM;
#endif

	if ((ntohl(h->flag) & 1)) {
		gtask_delete_htxf(sess_from_htlc(htlc), htxf);
		xfer_delete(htxf);
		return;
	}

	dh_start(htlc)
		switch (ntohs(dh->type)) {
		case HTLS_DATA_HTXF_REF:
			dh_getint(ref);
			break;
		case HTLS_DATA_QUEUE:
			dh_getint(queue);
			break;
		case HTLS_DATA_RFLT:
			if (ntohs(dh->len) >= 66) {
				L32NTOH(data_pos, &dh->data[46]);
				L32NTOH(rsrc_pos, &dh->data[62]);
			}
			break;
		}
	dh_end()


	if (!ref)
		return;
	htxf->data_pos = data_pos;
	htxf->rsrc_pos = rsrc_pos;
	htxf->queue = queue;
	if (!stat(htxf->path, &sb))
		htxf->data_size = sb.st_size;
	htxf->rsrc_size = resource_len(htxf->path);
	htxf->total_size = 133 + ((htxf->rsrc_size - htxf->rsrc_pos) ? 16 : 0) + comment_len(htxf->path)
			 + (htxf->data_size - htxf->data_pos) + (htxf->rsrc_size - htxf->rsrc_pos)+1;
	htxf->ref = ref;
	gettimeofday(&htxf->start, 0);

#ifdef USE_IPV6
	sprintf(portstr, "%u", server_port+1);
	if((error = getaddrinfo(htlc->addr->ai_canonname, portstr, &hints,

							&htxf->listen_addr))) {
		hx_printf_prefix(&sessions[0].htlc, 0, INFOPREFIX, "htxf %s:%u failed: %s\n", htlc->addr->ai_canonname, server_port+1, gai_strerror(error));
		gtask_delete_htxf(sess_from_htlc(htlc), htxf);
		xfer_delete(htxf);
		return;
	}
#else
	htxf->listen_addr = htlc->addr;
	htxf->listen_addr.sin_port = htons(ntohs(htxf->listen_addr.sin_port)+1);
#endif

	hx_output.xfer_queue(htxf);
	xfer_ready_write(htxf);
}
